import numpy as np
from datetime import timedelta
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.colors as colors

from pyadlml.stats import device_state_fraction
from pyadlml.constants import DEVICE, TIME, VALUE, BOOL, CAT, START_TIME, END_TIME, NUM
from pyadlml.dataset.stats.devices import state_cross_correlation as stat_scc, \
    inter_event_intervals as stat_iei, event_cross_correlogram as stat_event_cc, events_one_day, \
    event_count as stats_event_count, event_cross_correlogram2

from .util import heatmap_square, func_formatter_seconds2time, \
    heatmap, annotate_heatmap, savefig, _num_bars_2_figsize, save_fig, \
    _num_items_2_heatmap_square_figsize, _num_boxes_2_figsize, \
    _num_items_2_heatmap_one_day_figsize, _num_items_2_heatmap_square_figsize_ver2, LOG, xaxis_format_one_day, \
    sort_devices, add_top_axis_time_format, map_time_to_numeric, plot_cc, create_todo, xaxis_format_time, \
    dev_raster_data_gen, xaxis_format_time2, get_qualitative_cmap, plot_grid

from pyadlml.dataset.plot.plotly.util import format_device_labels
from pyadlml.dataset._core.devices import _is_dev_rep2, device_events_to_states, split_devices_binary, contains_non_binary
from pyadlml.dataset.stats.util import comp_tds_sums, comp_tds_sums_mean, comp_tds_sums_median
from pyadlml.dataset.util import select_timespan, infer_dtypes, str_to_timestamp, device_order_by
from pyadlml.util import get_sequential_color, get_secondary_color, get_primary_color, get_diverging_color


@save_fig
def inter_event_intervals(df_devices=None, inter_event_intervals=None, scale='log',
                          nr_merged_events_at=[], n_bins=50, figsize=(10, 6),
                          color=None, file_path=None):
    """
    Plot a histogram of the differences between succeeding device triggers.

    Parameters
    ----------
    df_devices : pd.DataFrame, default=None
        Recorded devices from a dataset. Fore more information refer to the
        :ref:`user guide<device_dataframe>`.
    inter_event_intervals : ndarray, default=None
        Array time deltas used to plot the histogram. Compute this array using (TODO REF TO STATS)
    nr_merged_events_at : list of strings, default=[]
        Timepoints for wich the fraction of events disregarded due to
        value manipulation by discretizing time-series data is plotted
    n_bins : int, default=50
        The number of histogram bins.
    color : str, default=None
        sets the color of the plot. When not set, the primary theming color is used.
        Learn more about theming in the :ref:`user guide <theming>`
    figsize : (float, float), default: None
        Width and height in inches. If not provided, the figsize is inferred automatically.
    file_path : str, default=None
        If set, saves the plot under the given file path and returns *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_device_inter_event_intervals
    >>> plot_device_inter_event_intervals(df_devices=data.df_devices)

    .. image:: ../_static/images/plots/dev_hist_trigger_td.png
       :height: 300px
       :width: 500 px
       :scale: 100 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing.


    """
    assert not (df_devices is None and inter_event_intervals is None)
    title='Device Inter-event-times'
    log_sec_col = 'total_log_secs'
    sec_col = 'total_secs'
    ylabel='count'
    xlabel_top = 'time'
    ax1label = '$\Delta t$'
    ax3label = 'imputed events'
    xlabel = 'log seconds' if scale == 'log' else 'seconds'
    color = (get_primary_color() if color is None else color)
    color2 = get_secondary_color()

    if inter_event_intervals is None:
        X = stat_iei(df_devices.copy())
    else:
        X = inter_event_intervals


    if nr_merged_events_at:
        from pyadlml.dataset.stats.devices import resampling_imputation_loss
        imp_frac_y = np.zeros(len(nr_merged_events_at))
        for i, dt in enumerate(nr_merged_events_at):
            imp_frac_y[i] = resampling_imputation_loss(df_devices, dt, return_fraction=True)

        # compute corresponding time in seconds
        imp_frac_x = [pd.Timedelta(dt).seconds for i, dt in enumerate(nr_merged_events_at)]

    # make equal bin size from max to min
    bins = np.logspace(min(np.log10(X)), max(np.log10(X)), n_bins)

    # make data ready for hist
    hist, _ = np.histogram(X, bins=bins)

    # plots
    fig, ax = plt.subplots(figsize=figsize)
    if scale == 'log':
        plt.xscale(LOG)
    ax.hist(X, bins=bins, label=ax1label, color=color)
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)

    if nr_merged_events_at:
        # create axis for line
        ax2=ax.twinx()
        #ax2.plot(bins, cum_percentage, 'r', label=ax2label, color=color2)
        ax2.set_ylabel('%')
        ax2.set_xscale(LOG)
        ax2.plot(imp_frac_x, imp_frac_y, label=ax3label, color=color2, marker='.')
        ax2.set_ylim([0, 1])

    add_top_axis_time_format(ax)

    # plot single legend for multiple axis
    h1, l1 = ax.get_legend_handles_labels()

    if nr_merged_events_at:
        h2, l2 = ax2.get_legend_handles_labels()
        ax.legend(h1+h2, l1+l2, loc='center right')
    else:
        ax.legend(h1, l1, loc='center right')
    plt.title(title, y=1.08)
    
    return fig


@save_fig
def state_boxplot(df_devs, binary_state='on', categories=False, order='mean',
                  scale='log', figsize=None, file_path=None):
    """
    generates a boxplot for all devices.

    Parameters
    ----------
    df_devs : pd.DataFrame, optional
        recorded devices from a dataset. Fore more information refer to the
        :ref:`user guide<device_dataframe>`.
    figsize : (int, int), default=None
        Width, height in inches. If not provided, the figsize is inferred automatically.
    binary_state : str one of {"on", "off"}, default='on'
        Determines which of the state is choosen for a boxplot for binary devices
    categories : bool, default=False
        If set a boxplot for each categorie of a categorical device is drawn
    order : {'mean', 'alphabetically', 'room'}, default='mean'
        determines the order in which the devices are listed.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.dataset import fetch_amsterdam
    >>> data = fetch_amsterdam()

    >>> from pyadlml.plot import plot_device_boxplot
    >>> plot_device_boxplot(data.df_devices)

    .. image:: ../_static/images/plots/dev_bp_dur.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing.
    """
    title = 'Device state distribution'
    xlabel = 'log seconds'
    xlabel_top = 'time'
    from pyadlml.dataset.stats.devices import state_times

    df = state_times(df_devs, binary_state=binary_state, categorical=categories)

    # select data for each device
    devices = list(df[DEVICE].unique())
    num_dev = len(devices)
    figsize = (_num_boxes_2_figsize(num_dev) if figsize is None else figsize)

    dat = np.zeros((num_dev,), dtype=object)
    for i, device in enumerate(devices):
        dat[i] = df.loc[df[DEVICE] == device, 'td'].dt.seconds.to_numpy()


    new_devices, new_order = format_device_labels(devices, infer_dtypes(df_devs), categorical_state=True)
    devices = np.flip(new_devices) # boxplots plot in reverse order
    dat = dat[np.flip(new_order)]

    # plotting
    fig, ax = plt.subplots(figsize=figsize)
    ax.boxplot(dat, vert=False)
    ax.set_title(title)
    ax.set_yticklabels(devices, ha='right')
    ax.set_xlabel(xlabel)
    ax.set_xscale(LOG)

    # create secondary axis with time format 1s, 1m, 1d
    add_top_axis_time_format(ax, xlabel_top)

    return fig


@save_fig
def event_density_one_day(df_devices=None, df_tod=None, dt='1h',
                          figsize=None, cmap=None, file_path=None):
    """
    Plots the heatmap for one day where all the device triggers are shown

    Parameters
    ----------
    df_devices : pd.DataFrame, optional
        recorded devices from a dataset. Fore more information refer to the
        :ref:`user guide<device_dataframe>`.
    df_tod : pd.DataFrame
        A precomputed transition table. If the *df_trans* parameter is given, parameters
        *df_acts* and *lst_acts* are ignored. The transition table can be computed
        in :ref:`stats <stats_acts_trans>`.
    dt : str of {'[1-12]h', default='1h'
        the resolution, time_bins in hours. The number are
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    cmap : str or Colormap, optional
        The Colormap instance or registered colormap name used to map scalar
        data to colors. This parameter is ignored for RGB(A) data.
        Defaults 'viridis'.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_device_event_density
    >>> plot_device_event_density(data.df_devices, dt='1h')

    .. image:: ../_static/images/plots/dev_hm_trigger_one_day.png
       :height: 300px
       :width: 500 px
       :scale: 100 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing.
    """
    assert not (df_devices is None and df_tod is None)
    title = "Device event density"
    xlabel = 'time'
    cbarlabel = 'counts'

    if df_tod is None:
        df = events_one_day(df_devices.copy(), dt=dt)
    else:
        df = df_tod

    if dt is None:
        raise NotImplementedError
    else:
        num_dev = len(list(df.columns))
        figsize = (_num_items_2_heatmap_one_day_figsize(num_dev) if figsize is None else figsize)
        cmap = (get_sequential_color() if cmap is None else cmap)

        x_labels = list(df[TIME])
        devs = df.columns[1:]
        dat = df.iloc[:, 1:].values.T

        # reorder devices
        devs, new_order = format_device_labels(devs, infer_dtypes(df_devices))
        dat = dat[new_order, :]

        # begin plotting
        fig, ax = plt.subplots(figsize=figsize)
        im, cbar = heatmap(dat, devs, x_labels, ax=ax, cmap=cmap, cbarlabel=cbarlabel)
        ax.set_title(title)
        ax.set_xlabel(xlabel)
        xaxis_format_one_day(ax, *ax.get_xlim())

        ax.set_aspect(aspect='auto')

    return fig


@save_fig
def event_correlogram(df_devs=None, lst_devs=None, df_tcorr=None, t_window='5s', figsize=None,
                             z_scale="linear", cmap=None, numbers=None, file_path=None):
    """
    Plot todo

    Parameters
    ----------
    df_devs : pd.DataFrame, optional
        recorded devices from a dataset. Fore more information refer to the
        :ref:`user guide<device_dataframe>`.
    lst_devs : lst of str, optional
        A list of devices that are included in the statistic. The list can be a
        subset of the recorded activities or contain activities that are not recorded.
    df_tcorr : pd.DataFrame
        A precomputed correlation table. If the *df_tcorr* parameter is given, parameters
        *df_devs* and *lst_devs* are ignored. The transition table can be computed
        in :ref:`stats <stats_devs_tcorr>`.
    t_window : str of {'[1-12]h', default='1h'
        the resolution, time_bins in hours. The number are
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    z_scale : {"log", "linear"}, default: None
        The axis scale type to apply.
    numbers : bool, default: True
        Whether to display numbers inside the heatmaps fields or not.
    cmap : str or Colormap, optional
        The Colormap instance or registered colormap name used to map scalar
        data to colors. This parameter is ignored for RGB(A) data.
        Defaults 'viridis'.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_device_event_correlogram
    >>> plot_device_event_correlogram(data.df_devices, dt='1h')

    .. image:: ../_static/images/plots/dev_hm_trigger_one_day.png
       :height: 300px
       :width: 500 px
       :scale: 100 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing.
    """
    assert not (df_devs is None and df_tcorr is None)

    title = "Triggercount with sliding window of " + t_window
    color = 'trigger count'
    cbarlabel = 'counts'
    log = True if z_scale == 'log' else False
    valfmt = "{x:.0f}"


    if df_tcorr is None:
        df = event_cross_correlogram(df_devs, lst_devs=lst_devs, t_window=t_window)
    else:
        df = df_tcorr

    # get the list of cross tabulations per t_window
    vals = df.astype(int).values.T
    devs = list(df.index)

    num_dev = len(devs)
    figsize = (_num_items_2_heatmap_square_figsize_ver2(num_dev) if figsize is None else figsize)
    cmap = (get_sequential_color() if cmap is None else cmap)

    fig, ax = plt.subplots(figsize=figsize)
    im, cbar = heatmap_square(vals, devs, devs, ax=ax, cmap=cmap,
                        cbarlabel=cbarlabel, log=log)#, cbar_kw=cbar_kw)
    
    # show numbers for small sizes
    if numbers is None: 
        if num_dev < 20:
            texts = annotate_heatmap(im, textcolors=("white", "black"), log=log, valfmt=valfmt)
    elif numbers:
        texts = annotate_heatmap(im, textcolors=("white", "black"), log=log, valfmt=valfmt)

    ax.set_title(title)
    fig.tight_layout()

    return fig


@save_fig
def state_similarity(df_devs=None, df_state_sim=None, figsize=None, order='alphabetical',
                     numbers=None, file_path=None):
    """
    Plots the cross correlation between the device signals

    Parameters
    ----------
    df_devs : pd.DataFrame, optional
        recorded devices from a dataset. Fore more information refer to the
        :ref:`user guide<device_dataframe>`.
    df_state_sim : pd.DataFrame
        A precomputed correlation table. If the *df_tcorr* parameter is given, parameters
        *df_devs* and *lst_devs* are ignored. The transition table can be computed
        in :ref:`stats <stats_devs_tcorr>`.
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    numbers : bool, default: True
        Whether to display numbers inside the heatmaps fields or not.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_device_state_similarity
    >>> plot_device_state_similarity(data.df_devs)

    .. image:: ../_static/images/plots/dev_hm_dur_cor.png
       :height: 400px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center


    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing.
    """
    assert not (df_devs is None and df_state_sim is None)

    title = 'State similarity'
    cmap = 'RdBu'
    cbarlabel = 'similarity'
    
    if df_state_sim is None:
        ct = stat_scc(df_devs)
    else:
        ct = df_state_sim
    ct = ct.replace(pd.NA, np.inf)
    values = ct.values.T
    devs = ct.index

    devs, new_order = format_device_labels(devs, infer_dtypes(df_devs), categorical_state=True)
    values = values[:, new_order]
    values = values[new_order, :]

    num_dev = len(devs)
    figsize = (_num_items_2_heatmap_square_figsize_ver2(num_dev) if figsize is None else figsize)
    fig, ax = plt.subplots(figsize=figsize)
    im, cbar = heatmap_square(values, devs, devs, ax=ax, cmap=cmap, cbarlabel=cbarlabel,
                       vmin=-1, vmax=1)
    if numbers is None:
        if num_dev < 15:
            valfmt = "{x:.2f}"
            texts = annotate_heatmap(im, textcolors=("black", "white"), 
                                     threshold=0.5, valfmt=valfmt)
        elif num_dev < 30:
            valfmt = "{x:.1f}"
            texts = annotate_heatmap(im, textcolors=("black", "white"), 
                                     threshold=0.5, valfmt=valfmt)
    if numbers:
        texts = annotate_heatmap(im, textcolors=("black", "white"), 
                                 threshold=0.5, valfmt="{x:.2f}")
    ax.set_title(title)
    fig.tight_layout()

    return fig

@save_fig
def state_fractions(df_devs=None, df_states=None, figsize=None,
                    color=None, color_sec=None, order='frac_on', file_path=None):
    """
    Plot bars the on/off fraction of all devices

    Parameters
    ----------
    df_devs : pd.DataFrame, optional
        recorded devices from a dataset. Fore more information refer to the
        :ref:`user guide<device_dataframe>`.
    df_states : pd.DataFrame
        A precomputed correlation table. If the *df_tcorr* parameter is given, parameters
        *df_devs* and *lst_devs* are ignored. The transition table can be computed
        in :ref:`stats <stats_devs_tcorr>`.
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    color : str, optional
        sets the primary color of the plot. When not set, the primary theming color is used.
        Learn more about theming in the :ref:`user guide <theming>`
    color_sec : str, optional
        sets the secondary color of the plot. When not set, the secondary theming color is used.
        Learn more about theming in the :ref:`user guide <theming>`
    order : {'frac_on', 'alphabetically', 'area'}, default='frac_on'
        determines the order in which the devices are listed.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_device_state_fractions
    >>> plot_device_state_fractions(data.df_devices)

    .. image:: ../_static/images/plots/dev_on_off.png
       :height: 300px
       :width: 500 px
       :scale: 100 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing.
    """
    assert not (df_devs is None and df_states is None)
    assert order in ['frac_on', 'alphabetical', 'area']

    title = 'Device state fractions'
    xlabel ='fraction per state'
    ylabel = 'Devices'
    on_label = 'on'
    off_label = 'off'

    color = (get_primary_color() if color is None else color)
    color2 = (get_secondary_color()if color_sec is None else color_sec)

    if df_states is None:
        df = device_state_fractions(df_devs)
    else:
        df = df_states

    num_dev = len(df_devs[DEVICE].unique())
    figsize = (_num_bars_2_figsize(num_dev) if figsize is None else figsize)

    df = sort_devices(df, 'frac')

    from pyadlml.dataset.util import infer_dtypes
    dtypes = infer_dtypes(df_devs)


    # Figure Size 
    fig, ax = plt.subplots(figsize=figsize)
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    # plot boolean devices
    df_bool = df[df[DEVICE].isin(dtypes[BOOL])].copy().sort_values(by=DEVICE)
    off_values = df_bool.loc[(df_bool[VALUE] == False), 'frac'].values
    ax.barh(dtypes[BOOL], off_values, label=off_label, color=color)
    # careful: notice "bottom" parameter became "left"
    ax.barh(dtypes[BOOL], (1-off_values), left=off_values, label=on_label, color=color2)

    # set the text centers for the boolean devices to the middle for the greater fraction
    first_number_left = True
    for i, dev in enumerate(dtypes[BOOL]):
        text_in_off = (off_values[i] >= 0.5)
        center = off_values[i]/2 if text_in_off else (1 - off_values[i])/2 + off_values[i]
        text = off_values[i] if text_in_off else (1 - off_values[i])
        if i == len(dtypes[BOOL])-1 and text_in_off:
           first_number_left = False
        text_color = 'white' if text_in_off else 'black'
        ax.text(center, i, '{:.4f}'.format(text), ha='center', va='center', color=text_color)

    # plot categorical devices
    cat_centers = []
    for j, dev in enumerate(dtypes[CAT]):
        categories = df.loc[(df[DEVICE] == dev), VALUE].copy().sort_values().values
        cum_sum = 0
        cat_centers.append([])
        for i, cat in enumerate(categories):
            value = df.loc[(df[DEVICE] == dev) & (df[VALUE] == cat), 'frac'].values[0]
            # save the center in bar plot as well as the value for setting
            # the annotations later
            cat_centers[j].append((cum_sum + value/2, value))
            ax.barh([dev], [value], label=dev + ' : ' + cat, left=[cum_sum])
            cum_sum += value


    # set the text centers for categorical devices into the middle for each category
    for c, y in zip(range(len(dtypes[CAT])), range(len(dtypes[BOOL]), num_dev-1)):
        for center, text in cat_centers[c]:
            ax.text(center, y, '{:.4f}'.format(text), ha='center', va='center', color='white')

    if dtypes[CAT]:
        ax.legend(bbox_to_anchor=(1, 1))
    else:
        if first_number_left:
            ax.legend(ncol=2, bbox_to_anchor=(0, 1),
                  loc='upper left', fontsize='small')
        else:
             ax.legend(ncol=2, bbox_to_anchor=(1,1),
                  loc='upper right', fontsize='small')


    # Remove axes splines 
    for s in ['top', 'right']: 
        ax.spines[s].set_visible(False)

    return fig


@save_fig
def event_count(df_devs=None, df_tc=None, figsize=None,
                scale='linear', color=None, order='count', file_path=None):
    """
    bar chart displaying how often activities are occurring

    Parameters
    ----------
    df_devs : pd.DataFrame, optional
        recorded devices from a dataset. Fore more information refer to the
        :ref:`user guide<device_dataframe>`.
    df_tc : pd.DataFrame
        A precomputed correlation table. If the *df_tcorr* parameter is given, parameters
        *df_devs* and *lst_devs* are ignored. The transition table can be computed
        in :ref:`stats <stats_devs_tcorr>`.
    scale : {"log", "linear"}, default: None
        The axis scale type to apply.
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    color : str, optional
        sets the primary color of the plot. When not set, the primary theming color is used.
        Learn more about theming in the :ref:`user guide <theming>`
    order : {'count', 'alphabetically', 'area'}, default='count'
        determines the order in which the devices are listed.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_device_event_count
    >>> plot_device_event_count(data.df_devices)

    .. image:: ../_static/images/plots/dev_bar_trigger.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing.
    """
    assert not (df_devs is None and df_tc is None)
    assert scale in ['log', 'linear']
    assert order in ['alphabetic', 'count', 'room']
    
    title = 'Device events'
    x_label = 'count'
    df_col = 'event_count'

    df = (stats_event_count(df_devs.copy()) if df_tc is None else df_tc)
    num_dev = len(df)
    figsize = (_num_bars_2_figsize(num_dev) if figsize is None else figsize)
    color = (get_primary_color() if color is None else color)

    # sort devices according to strategy
    sort_strategy = df_col if (order == 'count') else order
    df = sort_devices(df, sort_strategy)

    # plot
    fig, ax = plt.subplots(figsize=figsize)
    plt.title(title)
    plt.xlabel(x_label)
    ax.barh(df[DEVICE], df[df_col], color=color)

    if scale == LOG:
        ax.set_xscale(LOG)

    return fig


@save_fig
def event_raster(df_devices, figsize=(10, 6), start_time=None, end_time=None,
                 order='alphabetical', file_path=None):
    """ Plots an event raster

    """
    event_color = 'grey'
    title = 'Event raster'
    y_label = 'Devices'

    df_devs = df_devices.copy()\
                        .sort_values(by=TIME)\
                        .reset_index(drop=True)

    start_time = str_to_timestamp(start_time) if (start_time is not None) else None
    end_time = str_to_timestamp(end_time) if (end_time is not None) else None
    df_devs = select_timespan(df_devs=df_devs, start_time=start_time, end_time=end_time)

    if start_time is None:
        start_time = df_devs[TIME].iloc[0]
    if end_time is None:
        end_time = df_devs[TIME].iloc[-1]

    devs = device_order_by(df_devs, rule=order)

    df_devs['num_time'], start_time, end_time = map_time_to_numeric(
        df_devs[TIME], start_time=start_time, end_time=end_time)

    data_lst = dev_raster_data_gen(df_devs, devs)

    # create list of lists, where each list corresponds to an activity with tuples of start_time and time_length
    fig, ax = plt.subplots(figsize=figsize)
    ax.eventplot(data_lst, linelengths=[0.2]*len(devs), colors=event_color)

    ax.set_title(title)
    ax.set_yticks(np.arange(len(devs)))
    ax.set_yticklabels(devs.tolist())
    ax.set_ylabel(y_label)
    ax.set_xlim([0, 1])

    xaxis_format_time2(fig, ax, start_time, end_time)

    return fig


@save_fig
def plot_time_spans_slicer(X, window_size, stride, unit='m', file_path=None):
    unit = 'h'
    title = 'Time-span per window'
    y_label = 'counts'

    sums = comp_tds_sums(X, window_size, stride)
    mean = comp_tds_sums_mean(X, window_size, stride)
    median = comp_tds_sums_median(X, window_size, stride)

    seconds = np.zeros(shape=sums.shape)
    for i in range(len(sums)):
        seconds[i] = sums[i].seconds

    if unit == 's':
        f = 1
        x_label = 'seconds'
    elif unit == 'm':
        f = (1/60)
        x_label = 'minutes'
    else:
        f = (1/3600)
        x_label = 'hours'

    mean_str = str(int(mean.seconds*f))
    median_str = str(int(mean.seconds*f))

    hist, bin_edges = np.histogram(seconds*f, bins=20)
    fig = plt.figure()
    plt.hist(seconds*f, bins=bin_edges, label='timelength one window')
    plt.plot([mean.seconds*f]*2, [0, hist.max()], label='mean: ' + mean_str + ' ' + x_label)
    plt.plot([median.seconds*f]*2, [0, hist.max()], label='median: ' + median_str + ' ' + x_label)
    plt.legend()
    plt.title(title)
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    return fig


@save_fig
def event_cross_correlogram(df_devices=None, corr_data=(None, None, None), bin_size='1s', max_lag='2m', axis='off',
                            figsize=(5, 5), file_path=None, use_dask=False):
    #Plot cross-correlograms of all pairs.
    #   plotCCG(ccg,bins) plots a matrix of cross(auto)-correlograms for
    #   all pairs of clusters. Inputs are:
    #       ccg     array of cross correlograms           #bins x #clusters x #clusters
    #       bins    array with bin timings                #nbins x 0

    title = 'Event cross-correlogram'

    if corr_data[0] is None or corr_data[1] is None:
        ccg, bins = event_cross_correlogram2(df_devices, binsize=bin_size, maxlag=max_lag)
        devices = df_devices[DEVICE].unique()
    else:
        ccg = corr_data[0]
        bins = corr_data[1]
        devices = corr_data[2]

    fig = plot_cc(ccg, bins, title=title, y_label=devices, axis=axis, figsize=figsize, use_dask=use_dask)

    return fig


@save_fig
def states(df_devices, start_time=None,end_time=None, figsize=(20, 8),
           order='alphabetical', grid=False, file_path=""):
    """ Plots the devices events as raster combined with the activites over a timespan

    Parameters
    ----------
    """
    ylabel = 'Devices'
    title = 'Device states'
    binary_on_label = 'on'
    binary_off_label = 'off'

    color_dev_on = 'seagreen'
    color_dev_off = 'lightgrey'
    color_dev_num = 'blue'

    df_devs = df_devices.copy()\
                        .sort_values(by=TIME)\
                        .reset_index(drop=True)

    start_time = str_to_timestamp(start_time) if (start_time is not None) else None
    end_time = str_to_timestamp(end_time) if (end_time is not None) else None
    df_devs = select_timespan(df_devs=df_devs, start_time=start_time, end_time=end_time)

    if start_time is None:
        start_time = df_devs[TIME].iloc[0]
    if end_time is None:
        end_time = df_devs[TIME].iloc[-1]

    devs = df_devs[DEVICE].unique()

    fig, ax = plt.subplots(figsize=figsize)

    if grid:
        plot_grid(fig, ax, start_time, end_time)

    j = _plot_device_states(ax, df_devs, devs, start_time, end_time, color_dev_on, color_dev_off,
                           color_dev_num, binary_off_label, binary_on_label, return_nr_categories_used=True)

    ax.set_title(title)
    ax.set_yticks(np.arange(len(devs), dtype=np.float32))
    ax.set_yticklabels(devs.tolist())
    ax.set_ylabel(ylabel)
    ax.set_xlim([0, 1])

    handles, labels = ax.get_legend_handles_labels()
    ax.legend(handles, labels, bbox_to_anchor=(1, 1), loc="upper left")
    xaxis_format_time2(fig, ax, start_time, end_time)

    return fig

def _plot_device_states(ax, df_devs: pd.DataFrame, devs: list, start_time, end_time, color_on='red',
    color_off='grey', color_num='blue', off_label='off', on_label='on', return_nr_categories_used=False):
    """ Plot vertical bar plots for device each state

    Parameters
    ----------

    Returns
    -------
    j : int
        The number of colors used up for different devices.
    """
    col_bar_start = 'num_st'
    col_bar_len = 'diff'

    # do preprocessing
    dtypes = infer_dtypes(df_devs)
    df_devs = device_events_to_states(df_devs, extrapolate_states=True, start_time=start_time, end_time=end_time)

    df_devs['num_st'], _, _ = map_time_to_numeric(df_devs[START_TIME], start_time, end_time)
    df_devs['num_et'], _, _ = map_time_to_numeric(df_devs[END_TIME], start_time, end_time)
    df_devs['diff'] = df_devs['num_et'] - df_devs['num_st']


    j = 0
    tab = get_qualitative_cmap(len(devs))
    set_on_off_label = True
    for i, dev in enumerate(devs):
        df = df_devs[df_devs[DEVICE] == dev].copy()
        if dev in dtypes[BOOL]:
            tuples_off = df.loc[(df[VALUE] == False), [col_bar_start, col_bar_len]].values.tolist()
            tuples_on = df.loc[(df[VALUE] == True), [col_bar_start, col_bar_len]].values.tolist()

            ax.broken_barh(tuples_off, (i-0.25, 0.5), facecolors=color_off, label=off_label)
            ax.broken_barh(tuples_on, (i-0.25, 0.5), facecolors=color_on, label=on_label)

            #  The on and off label appear only once in the legend
            if set_on_off_label:
                set_on_off_label = False
                off_label, on_label = None, None

        elif dev in dtypes[CAT]:
            categories = df.loc[df[DEVICE] == dev, VALUE].unique()
            for cat in categories:
                values = df.loc[(df[VALUE] == cat), [col_bar_start, col_bar_len]].values.tolist()
                ax.broken_barh(values, (i-0.25, 0.5), facecolors=tab(j), label=dev + ' - ' + cat)
                j += 1

        elif dev in dtypes[NUM]:
            values = pd.to_numeric(df[VALUE])
            values = (values-values.min())/(values.max() - values.min())*0.5
            values = values + i - 0.25
            ax.plot(df['num_st'], values, color=color_num, linestyle='--', marker='o')

    if return_nr_categories_used:
        return j




import datetime
import unittest

from hassbrain_algorithm.models.hmm.hmm import ModelHMM_log_scaled
from hassbrain_algorithm.models.hmm.pchmm import PreConfHMM
from hassbrain_algorithm.controller import Controller
from hassbrain_algorithm.controller import Dataset
from hassbrain_algorithm.datasets.homeassistant import DatasetHomeassistant
from hbhmm.hmm.probs import Probs


class TestHomeassistantModelHMMLogScaled(unittest.TestCase):

    # Model part
    def setUp(self):
        # set of observations
        self.ctrl = Controller()
        self.ctrl.set_dataset(Dataset.HASS_TESTING)
        self.hass_obj = self.ctrl._dataset #type: DatasetHomeassistant
        self.hmm_model = PreConfHMM(self.ctrl)

    def tearDown(self):
        pass

    def test_load_custom_lists_modelHMM(self):
        custom_state_list = ['sleeping', 'cooking']
        custom_obs_list = [
            'binary_sensor.motion_bed',
            'binary_sensor.motion_mirror',
            'binary_sensor.motion_pc',
            'switch.test_switch_1',
            'light.test_light'
        ]
        hmm_model = self.hmm_model
        self.ctrl.set_custom_state_list(custom_state_list)
        self.ctrl.set_custom_obs_list(custom_obs_list)

        self.ctrl.load_dataset()
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)

    def test_load_modelHMM(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)
        print(self.hass_obj.get_obs_lbl_hashmap())
        print(self.hass_obj.get_state_lbl_hashmap())

    def test_train_modelHMM(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)

        self.ctrl.train_model()
        print(self.ctrl._model)

    def test_bench_modelHMM(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)

        self.ctrl.register_benchmark()
        self.ctrl.train_model()
        print(self.ctrl._model)
        report = self.ctrl.create_report(
            conf_matrix=True,
            accuracy=True,
            precision=True,
            recall=True,
            f1=True
        )
        print(report)

    def test_classify(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.train_model()
        hmm_model._hmm.set_format_full(True)
        print(hmm_model)
        print('-'*10)
        obs_seq = [('binary_sensor.motion_bed', 0), ('binary_sensor.motion_mirror', 1), ('binary_sensor.motion_bed', 0)]
        pred_state = hmm_model.classify(obs_seq)
        print('#'*100)
        print(pred_state)

    def test_classify_multi(self):
        """
        used to test for classification of multiple labels
        """
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.train_model()
        hmm_model._hmm.set_format_full(True)
        print(hmm_model)
        print('-'*10)
        obs_seq = [('binary_sensor.motion_bed', 0), ('binary_sensor.motion_mirror', 1)]#, ('binary_sensor.motion_bed', 0)]
        act_state_dict = hmm_model.classify_multi(obs_seq)
        print('#'*100)
        print(act_state_dict)
        #print(act_state_dict)
        #print(hmm_model.get_state_label_list())

    def test_pred_next_obs_single(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.train_model()
        hmm_model._hmm.set_format_full(True)
        #print(hmm_model)
        print('#'*100)
        obs_seq = [('binary_sensor.motion_bed', 0), ('binary_sensor.motion_mirror', 1), ('binary_sensor.motion_bed', 0)]
        tupel = hmm_model.predict_next_obs(obs_seq)
        print(tupel)

    def test_pred_next_obs_multi(self):
        hmm_model = self.hmm_model
        self.ctrl.load_dataset()
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.train_model()
        hmm_model._hmm.set_format_full(True)
        print(hmm_model)
        print('#'*100)
        obs_seq = [('binary_sensor.motion_bed', 0), ('binary_sensor.motion_mirror', 1), ('binary_sensor.motion_bed', 0)]
        #arr = hmm_model.predict_next_obs_arr(obs_seq)
        print(hmm_model._obs_lbl_hashmap)
        print(hmm_model._obs_lbl_rev_hashmap)
        res_dict = hmm_model.predict_prob_xnp1(obs_seq)
        print(hmm_model._obs_lbl_hashmap)
        print(hmm_model._obs_lbl_rev_hashmap)
        res_dict = hmm_model.predict_prob_xnp1(obs_seq)
        print(hmm_model._obs_lbl_hashmap)
        print(hmm_model._obs_lbl_rev_hashmap)
        print('#'*100)
        print(res_dict)


    def test_encode_loc_data(self):
        loc_data = [ {
            "name" : "loc1",
            "activities" : ['cooking'],
            "devices" : ['binary_sensor.motion_hallway', 'binary_sensor.motion_mirror'],
            },
            {"name" : "loc2",
            "activities" : ['cooking', 'eating'],
            "devices" : [],
            },
            {"name" : "loc3",
            "activities" : ['sleeping'],
            "devices" : ['binary_sensor.motion_bed'],
            },
        ]

        hmm_model = self.hmm_model
        self.ctrl.load_dataset()
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)

        #print('state_hm: ', hmm_model._state_lbl_hashmap)
        #print('obs_hm: ', hmm_model._obs_lbl_hashmap)

        #print('raw_loc_data: \t' + str(loc_data))
        enc_loc_data = hmm_model._encode_location_data(loc_data)
        #print('#'*100)
        #print('enc_loc_data: \t' + str(enc_loc_data))

    def test_encode_act_data(self):
        act_data = [
            {"name" : "cooking",
            "day_of_week" : 2,
            "start" : datetime.time.fromisoformat("06:15:00"),
            "end" : datetime.time.fromisoformat("08:45:00")
            },
            {"name" : "eating",
            "day_of_week" : 1,
            "start" : datetime.time.fromisoformat("06:15:00"),
            "end" : datetime.time.fromisoformat("08:45:00")
            },
            {"name" : "eating",
            "day_of_week" : 1,
            "start" : datetime.time.fromisoformat("08:46:00"),
            "end" : datetime.time.fromisoformat("10:00:00")
            },
        ]

        hmm_model = self.hmm_model
        self.ctrl.load_dataset()
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        print('raw_act_data: \t' + str(act_data))
        print('state_hm: ', hmm_model._state_lbl_hashmap)
        print('obs_hm: ', hmm_model._obs_lbl_hashmap)
        print('#'*100)
        enc_act_data = hmm_model._encode_act_data(act_data)
        print('enc_act_data: \t' + str(enc_act_data))


    def test_init(self):
        loc_data = [ {
            "name" : "loc1",
            "activities" : ['cooking'],
            "devices" : ['binary_sensor.motion_hallway', 'binary_sensor.motion_mirror'],
        },
            {"name" : "loc2",
             "activities" : ['cooking', 'eating'],
             "devices" : [],
             },
            {"name" : "loc3",
             "activities" : ['sleeping'],
             "devices" : ['binary_sensor.motion_bed'],
             },
        ]

        act_data = [
            {"name" : "sleeping",
            "day_of_week" : 0,
            "start" : datetime.time.fromisoformat("04:00:00"),
            "end" : datetime.time.fromisoformat("06:15:00")
            },
            {"name" : "cooking",
            "day_of_week" : 0,
            "start" : datetime.time.fromisoformat("06:15:00"),
            "end" : datetime.time.fromisoformat("08:45:00")
            },
            {"name" : "eating",
            "day_of_week" : 0,
            "start" : datetime.time.fromisoformat("08:46:00"),
            "end" : datetime.time.fromisoformat("10:00:00")
            },
            {"name" : "sleeping",
            "day_of_week" : 0,
            "start" : datetime.time.fromisoformat("12:00:00"),
            "end" : datetime.time.fromisoformat("13:00:00")
            },
            {"name" : "sleeping",
            "day_of_week" : 1,
            "start" : datetime.time.fromisoformat("02:00:00"),
            "end" : datetime.time.fromisoformat("06:30:00")
            },
            {"name" : "cooking",
            "day_of_week" : 1,
            "start" : datetime.time.fromisoformat("12:00:00"),
            "end" : datetime.time.fromisoformat("13:00:00")
            },
            {"name" : "cooking",
            "day_of_week" : 2,
            "start" : datetime.time.fromisoformat("19:00:00"),
            "end" : datetime.time.fromisoformat("00:00:00")
            },
            {"name" : "cooking",
            "day_of_week" : 2,
            "start" : datetime.time.fromisoformat("23:00:00"),
            "end" : datetime.time.fromisoformat("00:00:00")
            },
            {"name" : "sleeping",
            "day_of_week" : 2,
            "start" : datetime.time.fromisoformat("00:00:00"),
            "end" : datetime.time.fromisoformat("03:00:00")
            },
        ]

        hmm_model = self.hmm_model
        self.ctrl.load_dataset()
        self.ctrl.register_model(hmm_model)
        self.ctrl.register_location_info(loc_data)
        print('raw_act_data: \t' + str(act_data))
        print('state_hm: ', hmm_model._state_lbl_hashmap)
        #print('obs_hm: ', hmm_model._obs_lbl_hashmap)
        self.ctrl.register_activity_info(act_data)
        #print('#'*100)
        #enc_act_data = hmm_model._encode_act_data(act_data)
        #print('enc_act_data: \t' + str(enc_act_data))
        self.ctrl.init_model_on_dataset()
        hmm = hmm_model._hmm
        hmm.set_format_full(True)

        self.assertAlmostEqual(1.0, Probs.prob_to_norm(hmm._pi.sum()), 6)
        self.assertTrue(hmm.verify_emission_matrix())
        self.assertTrue(hmm.verify_transition_matrix())
        #print(self.hmm_model)




import numpy as np
def seconds2timestr(x):
    """
    Parameters
    ----------
    """
    if x-60 < 0:
        return "{:.1f}s".format(x)
    elif x-3600 < 0:
        return "{:.1f}m".format(x/60)
    elif x-86400 < 0:
        return "{:.1f}h".format(x/3600)
    else:
        return "{:.1f}t".format(x/86400)


def fmt_seconds2time_log(x):
    """ gets a normal input and formats it to log
    """
    x = np.log(x)
    if x-60 < 0:
        return "{:.0f}s".format(x)
    elif x-3600 < 0:
        return "{:.0f}m".format(x/60)
    elif x-86400 < 0:
        return "{:.0f}h".format(x/3600)
    else:
        return "{:.0f}t".format(x/86400)


def fmt_seconds2time(x):
    if x-60 < 0:
        return "{:.1f}s".format(x)
    elif x-3600 < 0:
        return "{:.1f}m".format(x/60)
    elif x-86400 < 0:
        return "{:.1f}h".format(x/3600)
    else:
        return "{:.1f}t".format(x/86400)





import unittest

from hassbrain_algorithm.models.hmm.hmm import *
from hassbrain_algorithm.controller import Controller
from hassbrain_algorithm.controller import Dataset


class TestController(unittest.TestCase):
    def setUp(self):
        # set of observations
        self.ctrl = Controller() # type: Controller
        #self.hmm_model = ModelHMM(self.ctrl)
        #self.hmm_model = ModelHMM_scaled(self.ctrl)
        #self.hmm_model = ModelHMM_log(self.ctrl)
        self.hmm_model = ModelHMM_log_scaled(self.ctrl)

    def tearDown(self):
        pass


    def test_ctrl_get_bench_metrics(self):
        hmm_model = self.hmm_model
        dk = Dataset.KASTEREN
        self.ctrl.load_dataset(dk)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.register_benchmark()
        self.ctrl.register
        # render
        #dot = self.ctrl.render_model()
        #dot.render('test.gv', view=True)

        self.ctrl.train_model(True)
        print(self.ctrl.get_bench_metrics())

    def test_ctrl_train_seqs(self):
        hmm_model = self.hmm_model
        dk = Dataset.KASTEREN
        self.ctrl.set_dataset(dk)
        self.ctrl.load_dataset()
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.register_benchmark()
        # render
        #dot = self.ctrl.render_model()
        #dot.render('test.gv', view=True)

        self.ctrl.train_model()
        report = self.ctrl.create_report(
            conf_matrix=True,
            accuracy=True,
            precision=True,
            recall=True,
            f1=True)

        print(self.ctrl._model)
        print(report)
        #self.ctrl.show_plot()


    def test_ctrl_presentation(self):
        hmm_model = self.hmm_model
        dk = Dataset.KASTEREN
        self.ctrl.load_dataset(dk)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.register_benchmark()
        # render
        #dot = self.ctrl.render_model()
        #dot.render('test.gv', view=True)

        self.ctrl.train_model(True)
        report = self.ctrl.create_report(
            conf_matrix=True,
            accuracy=True,
            precision=True,
            recall=True,
            f1=True)

        print(self.ctrl._model)
        print(report)
        #self.ctrl.show_plot()

        # render
        #dot = self.ctrl.render_model()
        #dot.render('test.gv', view=True)

    def test_generate_visualization(self):
        hmm_model = self.hmm_model
        dk = Dataset.KASTEREN
        self.ctrl.set_dataset(dk)
        self.ctrl.load_dataset()
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print('state_label_hm: ', hmm_model._state_lbl_hashmap)
        print('state_label_rev_hm: ', hmm_model._state_lbl_rev_hashmap)
        self.ctrl.save_visualization_to_file('/home/cmeier/code/tmp/visualization.png')


    def test_bench_train_loss(self):
        hmm_model = self.hmm_model
        dk = Dataset.KASTEREN
        self.ctrl.set_dataset(dk)
        self.ctrl.load_dataset()
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.register_benchmark()
        self.ctrl.register_loss_file_path('/home/cmeier/code/tmp/kasteren/train_loss.log')
        self.ctrl.train_model()
        self.ctrl.save_model('/home/cmeier/code/tmp/kasteren/kasteren_model.joblib')


    def test_bench_reports_conf_matrix(self):
        hmm_model = self.hmm_model
        dk = Dataset.KASTEREN
        self.ctrl.load_dataset(dk)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.register_benchmark()
        self.ctrl.train_model()
        report = self.ctrl.create_report(
            conf_matrix=True)
        print(report)

    def test_bench_reports(self):
        hmm_model = self.hmm_model
        dk = Dataset.KASTEREN
        self.ctrl.load_dataset(dk)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.register_benchmark()
        self.ctrl.train_model()
        report = self.ctrl.create_report(
            conf_matrix=True,
            accuracy=True,
            precision=True,
            recall=True,
            f1=True)
        print(report)
        #self.ctrl.show_plot()


    def test_bench_q_fct(self):
        hmm_model = self.hmm_model
        dk = Dataset.KASTEREN
        self.ctrl.load_dataset(dk)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.register_benchmark()
        print(self.ctrl._model)
        # use dataset Kasteren and q_fct
        self.ctrl.train_model(True)
        report = self.ctrl.create_report()
        print(self.ctrl._model)
        print(report)
        #self.ctrl.show_plot()

    def test_om(self):
        #plt.figure(figsize=(10,6))
        #self.pom.plot()
        hmm_model = self.hmm_model
        dk = Dataset.KASTEREN
        self.ctrl.load_dataset(dk)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        dot = self.ctrl.render_model()
        dot.render('test.gv', view=True)


    def test_train_model(self):
        hmm_model = self.hmm_model
        dk = Dataset.KASTEREN
        self.ctrl.set_dataset(dk)
        self.ctrl.load_dataset()
        self.ctrl.register_model(hmm_model)

        self.ctrl.init_model_on_dataset()



        #print(hmm_model._hmm)
        self.ctrl.train_model()
        print('#'*200)
        print(hmm_model)
        hmm_model._hmm.set_str_exp(True)
        hmm_model._hmm.set_format_full(True)
        print(hmm_model)
        #print(hmm_model._hmm.verify_transition_matrix())
        #print(hmm_model._hmm.verify_emission_matrix())

        #self._bench._model.draw()
        #self.ctrl.register_benchmark()
        #report = self.ctrl.create_report(True, True, True, True, True)
        #print(report)
        #self._bench.show_plot()


    def test_save_model(self):
        hmm_model = self.hmm_model
        dk = Dataset.KASTEREN
        self.ctrl.load_dataset(dk)
        self.ctrl.register_model(hmm_model)

        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)


        self.ctrl.train_model()
        self.ctrl.save_model()

    def test_init_model(self):
        hmm_model = self.hmm_model
        dk = Dataset.KASTEREN
        self.ctrl.set_dataset(dk)
        self.ctrl.load_dataset()
        self.ctrl.register_model(hmm_model)

        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)


    def test_load_model(self):
        hmm_model = self.hmm_model
        dk = Dataset.KASTEREN
        self.ctrl.set_dataset(dk)
        self.ctrl.load_dataset()
        self.ctrl.register_model(hmm_model)


        self.ctrl.load_model()
        print(self.ctrl._model)




import functools
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
from matplotlib.dates import AutoDateLocator
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib
from scipy.stats.kde import gaussian_kde
from sklearn.preprocessing import minmax_scale
import copy
from matplotlib.colors import LogNorm
import pandas as pd
from pyadlml.constants import TIME, END_TIME, START_TIME, DEVICE, VALUE, BOOL, CAT
from pyadlml.dataset.util import unitsfromdaystart
from pyadlml.util import get_sequential_color, get_diverging_color
import matplotlib.ticker as ticker

LOG = 'log'

def func_formatter_log_1(x, pos):
    """ gets a normal input and formats it to log with 1 decimal
    """
    if x == 0.0:
        return ""
    else:
        x = np.log10(x)
        return "{:.1f}".format(x)


def func_formatter_seconds2time_log(x, pos):
    """ gets a normal input and formats it to log
    """
    x = np.log(x)
    if x-60 < 0:
        return "{:.0f}s".format(x)
    elif x-3600 < 0:
        return "{:.0f}m".format(x/60)
    elif x-86400 < 0:
        return "{:.0f}h".format(x/3600)
    else:
        return "{:.0f}t".format(x/86400)


def func_formatter_seconds2time(x, pos):
    if x-60 < 0:
        return "{:.1f}s".format(x)
    elif x-3600 < 0:
        return "{:.1f}m".format(x/60)
    elif x-86400 < 0:
        return "{:.1f}h".format(x/3600)
    else:
        return "{:.1f}t".format(x/86400)


def heatmap(data, row_labels, col_labels, log=False, ax=None, cmap=None,
            cbar_kw={}, cbarlabel="", labels_top=False, pairgrid=False, colorbar=True,
            **kwargs):
    """
    Create a heatmap from a numpy array and two lists of labels.

    Parameters
    ----------
    data
        A 2D numpy array of shape (N, M).
    row_labels
        A list or array of length N with the labels for the rows.
    col_labels
        A list or array of length M with the labels for the columns.
    ax
        A `matplotlib.axes.Axes` instance to which the heatmap is plotted.  If
        not provided, use current axes or create a new one.  Optional.
    cbar_kw
        A dictionary with arguments to `matplotlib.Figure.colorbar`.  Optional.
    cbarlabel
        The label for the colorbar.  Optional.
    label_pos : str
        either bottom 
    **kwargs
        All other arguments are forwarded to `imshow`.
    """

    if not ax:
        ax = plt.gca()
    if cmap is None:
        cmap = get_diverging_color()
        plt.set_cmap(cmap)

    # We want to show all ticks...
    ax.set_yticks(np.arange(data.shape[0]))
    # ... and label them with the respective list entries.
    ax.set_yticklabels(row_labels)


    # Rotate the tick labels and set their alignment.
    ax.set_xticks(np.arange(data.shape[1]))
    ax.set_xticklabels(col_labels)
    if labels_top:
        ax.xaxis.tick_top()
        plt.setp(ax.get_xticklabels(), rotation=-45, ha="right",
            rotation_mode="anchor")
    else:
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
                rotation_mode="anchor")
    if pairgrid:
        # create grid for heatmap into every pair
        tcks = np.arange((data.shape[1])/2)*2 + 1.5
        ax.set_xticks(tcks, minor=True)
        ax.grid(which="minor", color="w", linestyle='-', linewidth=1)
        ax.tick_params(which="minor", bottom=False, left=False) 

    #if labels_top is not None:
    #    ax_top = ax.twiny()
    #    ax_top.set_xlim(0, data.shape[1])
    #    xticks = np.arange(data.shape[1]) + 0.5
    #    #print('xticks: ', xticks)
    #    ax_top.set_xticks(xticks)
    #    ax_top.set_xticklabels(labels_top)

    #    ax_top.plot([],[])

    #if h_sep is not None:
    #    # create a horziontal white seperating bar in the middle
    #    tcks = np.array([data.shape[0]/2 - 0.5])
    #    ax.set_yticks(tcks, minor=True)
    #    ax.grid(which="minor", color="w", linestyle='-', linewidth=4)
    #    ax.tick_params(which="minor", bottom=False, left=False)

    # Plot the heatmap
    if log:
        #set colorlabel for bad colors
        cmap = copy.copy(matplotlib.cm.get_cmap(cmap))
        cmap.set_bad(color = 'lightgrey', alpha = .8)
        
        # round to the next highest digit for display purposes
        vmax = 10**int(np.ceil(np.log10(data.max())))
        vmin = 1
        
        im = ax.imshow(data, cmap=cmap, norm=LogNorm(vmin=vmin, vmax=vmax),**kwargs)
    else:
        im = ax.imshow(data, **kwargs)
    ax.set_aspect('auto')

    # Create colorbar
    if colorbar:
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size='5%', pad=0.1)
        cbar = cax.figure.colorbar(im, cax, **cbar_kw) 
        cax.set_ylabel(cbarlabel, rotation=-90, va="bottom")

    return im, 0#cbar


def heatmap_square(data, row_labels, col_labels, log=False, ax=None, cmap=None,
            cbar_kw={}, cbarlabel="", grid=True, **kwargs):
    """
    Create a heatmap from a numpy array and two lists of labels.

    Parameters
    ----------
    data
        A 2D numpy array of shape (N, M).
    row_labels
        A list or array of length N with the labels for the rows.
    col_labels
        A list or array of length M with the labels for the columns.
    ax
        A `matplotlib.axes.Axes` instance to which the heatmap is plotted.  If
        not provided, use current axes or create a new one.  Optional.
    cbar_kw
        A dictionary with arguments to `matplotlib.Figure.colorbar`.  Optional.
    cbarlabel
        The label for the colorbar.  Optional.
    **kwargs
        All other arguments are forwarded to `imshow`.
    """

    if not ax:
        ax = plt.gca()

    if cmap is None:
        cmap = get_diverging_color()
        plt.set_cmap(cmap)

    # Plot the heatmap
    if log:
        cbarlabel = "log " + cbarlabel
        
        #set colorlabel for bad colors
        cmap = copy.copy(matplotlib.cm.get_cmap(cmap))
        cmap.set_bad(color = 'lightgrey', alpha = .8)
        
        # round to the next highest digit for display purposes
        vmax = 10**int(np.ceil(np.log10(data.max())))
        vmin = 1
        
        im = ax.imshow(data, cmap=cmap, norm=LogNorm(vmin=vmin, vmax=vmax),**kwargs)
    else:
        im = ax.imshow(data, cmap=cmap, **kwargs)

        
    # Create colorbar
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size='5%', pad=0.05)
    cbar = cax.figure.colorbar(im, cax, **cbar_kw) # debug
    cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")

    # We want to show all ticks...
    ax.set_xticks(np.arange(data.shape[1]))
    ax.set_yticks(np.arange(data.shape[0]))
    # ... and label them with the respective list entries.
    ax.set_xticklabels(col_labels)
    ax.set_yticklabels(row_labels)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")
    if grid:
        # Turn spines off and create white grid.
        ax.set_xticks(np.arange(data.shape[1]+1)-.5, minor=True)
        ax.set_yticks(np.arange(data.shape[0]+1)-.5, minor=True)
        ax.grid(which="minor", color="w", linestyle='-', linewidth=1)
        ax.tick_params(which="minor", bottom=False, left=False)

    return im, cbar


def annotate_heatmap(im, data=None, valfmt="{}",
                     textcolors=("black", "white"),
                     threshold=None, log=False, **textkw):
    """
    A function to annotate a heatmap.

    Parameters
    ----------
    im
        The AxesImage to be labeled.
    data
        Data used to annotate.  If None, the image's data is used.  Optional.
    valfmt
        The format of the annotations inside the heatmap.  This should either
        use the string format method, e.g. "$ {x:.2f}", or be a
        `matplotlib.ticker.Formatter`.  Optional.
    textcolors
        A pair of colors.  The first is used for values below a threshold,
        the second for those above.  Optional.
    threshold
        Value in data units according to which the colors from textcolors are
        applied.  If None (the default) uses the middle of the colormap as
        separation.  Optional.
    **kwargs
        All other arguments are forwarded to each call to `text` used to create
        the text labels.
    """
    if not isinstance(data, (list, np.ndarray)):
        data = im.get_array()

    # normalize the threshold if e.g counts are given
    if threshold is not None and not (0.0 < threshold and threshold < 1.0):
        threshold = im.norm(threshold)
    else:
        threshold = 0.5

    # Set default alignment to center, but allow it to be
    # overwritten by textkw.
    kw = dict(horizontalalignment="center",
              verticalalignment="center")
    kw.update(textkw)

    # Get the formatter in case a string is supplied
    if log and not isinstance(valfmt, matplotlib.ticker.FuncFormatter):
        def func_formatter_tmp(x, p):
            with np.errstate(divide='ignore'):
                x = np.log10(x)
                if np.isinf(x):
                    #return "-$\infty$"
                    return ""
                else:
                    return "{:.2f}".format(x)
        
        format_func = lambda x, p: func_formatter_tmp(x, p)
        valfmt = matplotlib.ticker.FuncFormatter(format_func)
    elif isinstance(valfmt, str):
        valfmt = matplotlib.ticker.StrMethodFormatter(valfmt)
        
    diverging_lst = ['PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu',
            'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic']
    
    # Loop over the data and create a `Text` for each "pixel".
    # Change the text's color depending on the data.
    texts = []
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            # if the content is masked for e.g log values set the color to black
            if isinstance(im.norm(data[i, j]), np.ma.core.MaskedConstant):
                kw.update(color='black')
                
            # if list is diverging high values should have white and values
            # near zero should have black values
            elif im.cmap.name in diverging_lst:
                epsilon = 0.25
                cond = im.norm(data[i, j]) # maps data to [0,1]
                # defines span in between of 0.5=2*eps that leads to black color
                if cond <= threshold - epsilon or threshold + epsilon <= cond:
                    color = 'white'
                else:
                    color = 'black'
                kw.update(color=color)
            else:
                kw.update(color=textcolors[int(im.norm(data[i, j]) > threshold)])
            text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
            texts.append(text)

    return texts


def ridgeline(data, overlap=0, fill=True, labels=None, n_points=150, dist_scale=0.05):
    """
    Creates a standard ridgeline plot.

    data, list of lists.
    overlap, overlap between distributions. 1 max overlap, 0 no overlap.
    fill, matplotlib color to fill the distributions.
    n_points, number of points to evaluate each distribution function.
    labels, values to place on the y axis to describe the distributions.
    """
    if overlap > 1 or overlap < 0:
        raise ValueError('overlap must be in [0 1]')
    #xx = np.linspace(np.min(np.concatenate(data)),
    #                 np.max(np.concatenate(data)), n_points)
    xx = np.linspace(0.0, 86400.0, n_points)
    
    curves = []
    ys = []

    for i, d in enumerate(data):
        if d[0] == -1:
            # the case when there is no sample from this activity
            curve = np.zeros(shape=xx.shape)
        else:
            curve = gaussian_kde(d).pdf(xx)

        y = i*(1.0-overlap)
        ys.append(y)
        curve = minmax_scale(curve)*dist_scale
        if fill:
            plt.fill_between(xx, np.ones(n_points)*y,
                             curve+y, zorder=len(data)-i+1, color=fill)
        plt.plot(xx, curve+y, c='k', zorder=len(data)-i+1)

    if labels:
        plt.yticks(ys, labels)


def savefig(fig, file_path):
    """ saves figure to folder and if folder doesn't exist create one
    """

    fp = Path(file_path).resolve()
    folder = fp.parent
    if not folder.is_dir():
        folder.mkdir(parents=True, exist_ok=True)
    plt.savefig(fp)


""" --------------------------------------------------------------------------
these methods are for infering visual correct properties of matplotlib plots
"""
def _num_bars_2_figsize(n):
    """ uses linear regression model to infer adequate figsize
        from the number of bars in a bar plot
    Data used for training:
        x = [7,9,10,12,22,23,26]
        y = [[8,4],[9,5],[9,5],[7,5],[10,10],[10,9],[10,11]]
    Returns
    -------
    (w,h) : tuple
        the width and the height of the figure
    """
    if n <= 7:
        return (8,4)
    else:
        y1 = int(0.10937*n + 7.29688)
        y2 = int(0.36367*n + 1.33711)
    return y1, y2


def _num_boxes_2_figsize(n):
    """ uses linear regression model to infer adequate figsize
        from the number of boxes in a boxplot
    Data used for training:
        X = [ 7  9 11 22 23 26]
        y = [[8,4],[9,5],[10,6],[10,10],[10,10],[10,10],[10,11]]
    Returns
    -------
    (w,h) : tuple
        the width and the height of the figure
    """
    if n <= 7:
        return (8,4)
    else:
        y1 = 0.07662*n + 8.24853
        y2 = 0.36444*n + 1.71415
    return int(y1), int(y2)


def _num_items_2_heatmap_square_figsize(n):
    """ uses linear regression model to infer adequate figsize
        from the number of items 
    Data used for training:
        X = [4, 8, 9, 10, 11, 15, 22, 26]
        y = [[4,4],[5,5],[5,5],[6,6],[8,8],[8,8],[10,10],[11,11]]
    Parameters
    ----------
    n : int
        number of items
    Returns
    -------
    (w,h) : tuple
        the width and the height of the figure
    """
    w = 0.32626*n + 2.84282
    h = w
    return (int(w), int(h))


def _num_items_2_heatmap_square_figsize_ver2(n):
    """ uses linear regression model to infer adequate figsize
        from the number of items 
        the difference from version 1 is from very long names leading
        to a smaller figure
    Data used for training:
        X = [5,10,15,20,25,30,35,50,60]
        y = [[6,6],[8,8],[9,9],[11,11],[11,11],[12,12],[13,13],[14,14]]
    Parameters
    ----------
    n : int
        number of items
    Returns
    -------
    (w,h) : tuple
        the width and the height of the figure
    """
    w = 0.12940*n + 6.86068
    h = w
    return (int(w), int(h))


def _num_items_2_heatmap_one_day_figsize(n):
    """ uses linear regression model to infer adequate figsize
        from the number of items 
    Data used for training:
        X = [2,4,6,10,15,20,30,40,50,60]
        y = [[10,1],[10,2],[10,3],[10,4],[10,6],[10,8],[10,10],[10,12],[10,15],[10,17]]
    Parameters
    ----------
    n : int
        number of items
    Returns
    -------
    (w,h) : tuple
        the width and the height of the figure
    """
    w = 10
    h = 0.27082*n + 1.38153 
    return (int(w), int(h))


def _num_items_2_ridge_figsize(n):
    """ uses linear regression model to infer adequate figsize
        from the number of boxes in a boxplot
    Data used for training:
        X = [1, 3, 4, 6, 8, 11, 14, 16, 19, 22]
        y = [[10,1],[10,3],[10,4],[10,6],[10,8],[10,10],[10,10],[10,12],[10,13],[10,14]]
    Parameters
    ----------
    n : int
        number of items
    Returns
    -------
    (w,h) : tuple
        the width and the height of the figure
    """
    y1 = 10
    y2 = 1*n
    return int(y1), int(y2)


def _num_items_2_ridge_ylimit(n):
    """ uses linear regression model to infer adequate figsize
        from the number of boxes in a boxplot
    Data used for training:
        X = [1, 3, 4, 6, 8, 11, 14, 16, 19, 22, 24]
        y = [.15, 0.5, 0.6, 0.9, 1.18, 1.7, 2.1, 2.4, 2.85, 3.3, 3.7]
    Parameters
    ----------
    n : int
        number of items
    Returns
    -------
    (w,h) : tuple
        the width and the height of the figure
    """
    return 0.15134*n + 0.00076
 

def _contingency_hm_get_figsize(n_activities, n_devices):
    """ uses linear regression model to infer adequate figsize
        from the number of activities and devices
    Data used for training non-split:
        X = [[3,3],[5,3],[5,8],[3,10],[8,10],[8,14],[5,10],[8,20],[8,30],[10,10],[15,35],[20,20],[15,20],[10,5],[12,35],[5,15],[5,20],[10,20],[10,30]]
        y = [[8,2],[8,3],[8,3],[8,4], [8,5], [8,5], [9,5], [10,4],[13,4],[7,7],  [14,4], [10,8], [11,6], [10,6],[14,4], [10,6],[12,6],[10,6],[13,8]]
    
    data used for training split:
        X = [[20,36],[10,36],[4,36],[8,36],[8,45],[15,45],[20,45],[20,50],[10,50],[5,50],[8,50],[8,60],[13,60],[20,60],[20,70],[16,70],[11,70],[8,70],[5,70],[22,143],[22,144]]
        y = [[10,13],[10,8], [10,4],[10,7],[12,7],[12,10],[13,13],[13,13],[13,10],[13,4],[13,6],[13,6],[13,9],[13,12],[13,12],[13,12],[13,9],[13,6],[13,4],[20,10],[35,16]]

    Parameters
    ----------
    n_activities : int
    n_devices : int

    Returns
    -------
    (w,h) : tuple
        the width and the height of the figure
    """
    if _contingency_hm_split_plot(n_devices):
        w = -0.03406*n_activities + 0.20607*n_devices + 6.89947
        h = 0.21291*n_activities + -0.01205*n_devices + 3.40532
    else:
        w = 0.02551*n_activities + 0.14933*n_devices + -2.1431*np.log(n_activities) + 11.35675
        h = 0.54000*n_activities + -0.01080*n_devices + 2.74955
    return (int(w),int(h))


def _contingency_hm_split_plot(num_dev):
    return num_dev > 35


def _hm_cont_infer_valfmt(num_dev):
    if not _contingency_hm_split_plot(num_dev):
        if num_dev < 15:
            return "{x:.2f}"
        elif num_dev < 30:
            return "{x:.1f}"
        else:
            raise ValueError
    else:
        if num_dev < 40:
            return "{x:.2f}"
        elif num_dev < 60:
            return "{x:.1f}"
        else:
            raise ValueError


def _hm_cont_infer_cb_width(num_dev):
    if num_dev > 140: return 0.01


def heatmap_contingency(activities, devices, values, title, cbarlabel, \
    valfmt=None, textcolors=("white", "black"), z_scale=None, numbers=None, figsize=None):
    """ plots a contingency table between activities and devices
    Parameters
    ----------
    activities : 1d np.array n long
    devices : 1d np.array m long
    values : (n,m) np.array
    title : asdf
    """
    acts = activities
    devs = devices
    vals = values
    num_dev = len(devs)
    num_act = len(acts)

    log = (True if z_scale == 'log' else False)
    cmap = get_sequential_color()

    if _contingency_hm_split_plot(num_dev):
        if num_dev % 2 != 0:
            split = int(np.floor(num_dev/2))
        else: 
            split = int(num_dev/2)
        val1 = vals[:,:split]
        val2 = vals[:,split:]

        devs1 = devs[:split]
        devs2 = devs[split:]

        # zero padding for second
        if num_dev % 2 != 0:
            val2 = np.concatenate((val2,np.zeros((val1.shape[0],1))),axis=1)    
            devs2 = np.append(devs2, np.array([""], dtype=object),axis=0)

        figsize = (_contingency_hm_get_figsize(num_act, num_dev) if figsize is None else figsize)
        fig, (ax1, ax2) = plt.subplots(figsize=figsize, nrows=2)
        im, cbar = heatmap(val1, acts, devs1, ax=ax1, log=log, cbarlabel=cbarlabel, cmap=cmap,
            labels_top=True, colorbar=False
        )
        im2, cbar = heatmap(val2, acts, devs2, ax=ax2, log=log, cbarlabel=cbarlabel, cmap=cmap,
             colorbar=False
        )

        fig.subplots_adjust(right=0.8)
        w = 0.03
        h = 0.76
        cax = fig.add_axes([0.82, 0.12, w, h])
        cbar = cax.figure.colorbar(im2, cax)
        cax.set_ylabel(cbarlabel, rotation=-90, va="bottom")
        fig.subplots_adjust(hspace=0.05)
        #wspace = 0.2   # the amount of width reserved for blank space between subplots
        #hspace = 0.2   # the amount of height reserved for white space between subplots

    else:
        figsize = (_contingency_hm_get_figsize(num_act, num_dev) if figsize is None else figsize)
        fig, ax = plt.subplots(figsize=figsize)
        im, cbar = heatmap(vals, acts, devs, ax=ax, log=log, cbarlabel=cbarlabel, cmap=cmap)

    #print('na,nd -> fs: [{},{}],->[{},{}],'.format(num_act, num_dev, *figsize))

    # only plot numbers if it is true or None
    if valfmt is None and numbers != False:
        try:
            valfmt = _hm_cont_infer_valfmt(num_dev)
        except ValueError:
            numbers = False
    if numbers is None or numbers:
        try:
            annotate_heatmap(im, textcolors=textcolors,
                    threshold=0.5, valfmt=valfmt)
            if _contingency_hm_split_plot(num_dev):
                 annotate_heatmap(im2, textcolors=textcolors,
                                    threshold=0.5, valfmt=valfmt)
        except ValueError:
            pass

    fig.suptitle(title)
    #plt.gcf().subplots_adjust()
    return fig


def plot_cv_impact_parameter(cv, parameter, figsize=(10,15)):
    """
    Plot all parameter combinations w.r.t. to all expressions of the given he parameter

    Parameters
    ----------
    cv : sklearn grid search object
        Todo
    parameter : str
        A parameters string representation

    """
    param_values = cv.param_grid[parameter]
    params = cv.cv_results_['params']
    score = cv.cv_results_['mean_test_%s'%(cv.scoring[0])]

    # create a list for each parameter in param_values
    # where the rest withouth the parameter and the score are kept
    # the result is automatically sorted correctly
    pv_ticks = [[] for i in param_values]
    pv_score = [[] for i in param_values]
    for i in range(len(params)):
        for j, param in enumerate(param_values):
            if params[i][parameter] == param:
                r = dict(params[i])
                del r[parameter]
                pv_ticks[j].append(r)
                pv_score[j].append(score[i])

    # plot figure
    plt.figure(figsize=figsize)
    for i, pv in enumerate(param_values):
        plt.plot(pv_score[i], label=parameter + ': ' + str(pv))

    plt.xticks(np.arange(len(pv_ticks[0])), pv_ticks[0], rotation=90)
    plt.ylabel(cv.scoring[0])
    plt.tight_layout()
    plt.legend()
    plt.show()


def map_time_to_numeric(time_array, start_time=None, end_time=None):
    """
    Parameters
    ----------
    time_array : nd.array or pd.Series or pd.DataFrame
    start_time : pd.Timestamp

    end_time : pd.Timestamp


    Returns
    -------
    res : nd.array
        Array of transformed timestamps
    start_time : pd.Timestamp

    end_time : pd.Timestamp
    """
    if isinstance(time_array, pd.DataFrame):
        time_array = time_array[TIME].values
    if isinstance(time_array, pd.Series):
        time_array = time_array.values

    # get start and end_time
    start_time = time_array[0] if start_time is None else start_time
    end_time = time_array[-1] if end_time is None else end_time

    if isinstance(start_time, pd.Timestamp):
        start_time = start_time.to_numpy()
    if isinstance(end_time, pd.Timestamp):
        end_time = end_time.to_numpy()


    # map to values between [0,1]
    res = (time_array - start_time)/(end_time - start_time)

    return res, start_time, end_time


def get_qualitative_cmap(n, as_array=False):
    """ Returns a matplotlib colormap based on the number of given categories.
    Allows for up to 40 categories

    Parameter
    ---------
    n : int, required
        Number of categories

    Returns
    -------
    tab : matplotlib.Colormap,
        A qualitative color map
    """
    if n <= 8:
        tab = plt.get_cmap('Set2')
    elif n <= 12:
        tab = plt.get_cmap('Set3')
    elif n <= 20:
        tab = plt.get_cmap('tab20')
    else:
        import matplotlib.colors as mcolors
        colors1 = plt.get_cmap('tab20b')(np.arange(0, 20, 1))
        colors2 = plt.get_cmap('tab20c')(np.arange(0, 20, 1))

        # interleave colors instead of vstack to be better visual distinguishable
        colors = np.empty((colors1.shape[0] + colors2.shape[0], colors1.shape[1]), dtype=colors1.dtype)
        colors[0::2] = colors1
        colors[1::2] = colors2

        tab = mcolors.ListedColormap(colors, name='tab40')

    if as_array:
        return tab[np.arange(0, n, 1)]
    else:
        return tab


def xaxis_format_one_day(ax, upper_limit, lower_limit):
    """
    Format the x-axis with hours ranging from 01:00 - 23:00
    """
    def func(x,p):
        if True:
            if np.ceil(x/k) < 10:
                return '{}:00'.format(int(x/k)+1)
            else:
                return '{}:00'.format(int(x/k)+1)

    # calculate the tick positions
    k = (lower_limit-upper_limit)/24
    ticks_pos = np.arange(0,23)*k + (-0.5 + k)

    locator = ticker.FixedLocator(ticks_pos)
    formatter = ticker.FuncFormatter(func)
    ax.xaxis.set_major_formatter(formatter)
    ax.xaxis.set_major_locator(locator)


def sort_devices(df, strategy):
    """
    Parameters
    ----------
    df : pd.DataFrame
        Device dataframe
    strategy : str of {alphabetical, area, ???)
        The strategy with which the devices are sorted.
        ??? stands for the dynamic quantity (column-name) that
        contains numerical value. (e.g counts would sort after the most
        occuring device)

    """
    if strategy == 'alphabetical':
        df = df.sort_values(by=DEVICE, axis=0)
    elif strategy == 'area':
        raise NotImplementedError('room order will be implemented in the future')
    else:
        df = df.sort_values(by=strategy, axis=0)

    return df

def dev_raster_data_gen(df_devs: pd.DataFrame, devs):
    data_lst = []
    for dev in devs:
        dev_df = df_devs[df_devs[DEVICE] == dev]
        data_lst.append(dev_df['num_time'].values)
    return data_lst

def create_todo(df_devs: pd.DataFrame, devs, color_dev_on: str, color_dev_off: str):
    """ Creates the data for a event raster plot from devices
    """
    data_lst = []
    dev_colors = []

    for dev in devs:
        # select values for each device
        dev_df = df_devs[df_devs[DEVICE] == dev]
        data_lst.append(dev_df['num_time'].values)

        if not df_devs_binary.empty:
            if dev in df_devs_binary:
                # create on-off color coding for each device
                dev_df[VALUE].replace(True, color_dev_on, inplace=True)
                dev_df[VALUE].replace(False, color_dev_off, inplace=True)
                dev_colors.append(dev_df[VALUE].values)
            else:
                dev_colors.append(['grey'])
        else:
            # create on-off color coding for each device
            dev_df[VALUE].replace(True, color_dev_on, inplace=True)
            dev_df[VALUE].replace(False, color_dev_off, inplace=True)
            dev_colors.append(dev_df[VALUE].values)
    return data_lst, dev_colors


def xaxis_format_time2(fig, ax, start_time: np.datetime64, end_time: np.datetime64):
    """
    Correctly set xlabels
    """
    ax.set_xticks([])

    newax = fig.add_axes(ax.get_position(), frameon=False)
    newax.plot([start_time, end_time], [0, 0], alpha=0.0)
    newax.set_yticks([])

    from matplotlib.dates import SecondLocator, MinuteLocator, DayLocator, WeekdayLocator, \
                                 MonthLocator, \
                                 AutoDateFormatter, DateFormatter

    from matplotlib.dates import HOURS_PER_DAY, MINUTES_PER_DAY, SEC_PER_DAY, \
        MUSECONDS_PER_DAY, \
        DAYS_PER_MONTH, DAYS_PER_YEAR, DAYS_PER_WEEK

    from matplotlib.ticker import FuncFormatter
    import datetime
    class Func():
        """ Prints for the first label the whole date and for the rest the appended stuff
        """
        def __init__(self, formatting, freq):
            self.formatting = formatting
            self.freq = freq

        def __call__(self, x, pos):
         x = matplotlib.dates.num2date(x)
         if pos == 0 and (self.freq == 'hourly' or self.freq == '10minutes'):
             fmt = '%d.%m.%Y - %H:%M:%S'
         elif pos == 0:
             fmt = '%d.%m.%Y'
         elif x.strftime('%d:%m') == '01:01':
             # if silvester
             fmt = '%d.%m.%Y'

         elif x.time() == datetime.time(0, 0) and self.freq == 'hourly':
             # append for hourly ticks adt midnight the days
             fmt = '%d:%m - ' + self.formatting
         else:
             fmt = self.formatting
         return x.strftime(fmt)

    locator = AutoDateLocator()
    formatter = AutoDateFormatter(locator)
    # distance in days between major ticks maps to formatting
    formatter.scaled = {
            DAYS_PER_MONTH: FuncFormatter(Func('%d.%m', freq='monthly')),           # ticks with at least a month
            DAYS_PER_WEEK: FuncFormatter(Func('%dx.%m', freq='weekly')),            # ticks with at least a week
            1: FuncFormatter(Func('%d.%m', freq='daily')),                          # ticks with at least a day
            1 / HOURS_PER_DAY: FuncFormatter(Func('%H:%M', freq='hourly')),         # hourly ticks
            1 / MINUTES_PER_DAY: FuncFormatter(Func('%H:%M', freq='10minutes')),    # ticks are ten minutes (3h)
            1 / SEC_PER_DAY: '%H:%M:%S',                                            #
    }

    newax.xaxis.set_major_locator(locator)
    newax.xaxis.set_major_formatter(formatter)

    plt.setp(newax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

def xaxis_format_time(ax , start_time : np.datetime64, end_time : np.datetime64):
    """
    Correctly set xlabels
    """
    # get number of seconds
    diff = int((end_time - start_time)*1e-9)
    # lesser than one minute

    #padding = 0.05
    #ax.set_xlim(-padding, 1+padding)
    x_range = 1  # data is distributed between 0 and 1

    sec_per_min = 60
    sec_per_hour = 60*60
    sec_per_day = 60*60*24
    sec_per_week = 60*60*24*7
    sec_per_half_month = 60*60*24*15
    sec_per_month = 60*60*24*30
    sec_per_half_year = 60*60*24*30*6

    def get_stuff_done(tick_step, start_time, int_range):
        """
        """
        # compute offset
        td_fixpoint_to_st = unitsfromdaystart(start_time, unit='s')
        nr_ticks_to_first_tick = int(np.ceil(td_fixpoint_to_st*(1/tick_step)))
        offset = int(tick_step*nr_ticks_to_first_tick - td_fixpoint_to_st)

        nr_ticks = np.floor(int_range*(1/tick_step)).astype(int)  # nr of tick that fall into the interval
        tmp = np.full(nr_ticks, tick_step)
        tmp[0] = offset
        ticks = np.cumsum(tmp)

        # normalize
        norm = 1 / int_range     # e.g 3-satz x/diff = y/(1.05 - (-0.05))
        ticks = ticks*norm
        return ticks, nr_ticks_to_first_tick


    if diff <= 60:
        # generate label every 10 seconds
        ticks, nr_tick2first_tick = get_stuff_done(tick_step=10,int_range=diff,
                                                   start_time=start_time)
        ax.set_xticks(ticks)
        labels = ['{}0 sec'.format(i) for i in range(nr_tick2first_tick, 6)]
        ax.set_xticklabels(labels, rotation=-45)

    # lesser than one hour
    elif diff < sec_per_hour:
        # generate label every 10 minutes
        ticks, nr_tick2first_tick = get_stuff_done(tick_step=sec_per_min*10,int_range=diff,
                                                   start_time=start_time)
        ax.set_xticks(ticks)
        labels = ['{}0 min'.format(i) for i in range(nr_tick2first_tick, 6)]
        ax.set_xticklabels(labels, rotation=-45)

    # lesser than a day
    elif diff < sec_per_day:
        # generate label every two hour
        ticks, nr_tick2first_tick = get_stuff_done(tick_step=sec_per_hour*2, int_range=diff,
                                                   start_time=start_time)
        ticks = list(ticks)
        labels = ['{} h'.format(2*i) for i in range(nr_tick2first_tick, len(ticks)+1)]
        labels.insert(0, pd.Timestamp(start_time).date())
        ticks.insert(0, 0)

        ax.set_xticks(ticks)
        ax.set_xticklabels(labels, rotation=-45)

    # lesser than a week
    elif diff < sec_per_week:
        # generate label every day
        ticks, nr_tick2first_tick = get_stuff_done(tick_step=sec_per_day, int_range=diff,
                                                   start_time=start_time)
        ax.set_xticks(ticks)
        dates = pd.date_range(start_time + pd.Timedelta('1D'), periods=len(ticks))
        labels = [date.date() for date in dates]
        ax.set_xticklabels(labels, rotation=45)
        ax.xaxis.set_minor_locator(ticker.AutoMinorLocator(n=4))


    elif diff < sec_per_half_month:
        # generate label every second day
        ticks, nr_tick2first_tick = get_stuff_done(tick_step=sec_per_day*2, int_range=diff,
                                                   start_time=start_time)
        ticks = list(ticks)
        dates = pd.date_range(start_time + pd.Timedelta('2D'), periods=len(ticks))
        labels = [date.date() for date in dates]

        ax.set_xticks(ticks)
        ax.set_xticklabels(labels, rotation=45)
        ax.xaxis.set_minor_locator(ticker.AutoMinorLocator(n=2))

    elif diff < sec_per_month:
        # generate label every day
        ticks, nr_tick2first_tick = get_stuff_done(tick_step=sec_per_day*2, int_range=diff,
                                                   start_time=start_time)
        ticks = list(ticks)
        dates = pd.date_range(start_time + pd.Timedelta('2D'), periods=len(ticks))
        labels = [date.date() for date in dates]
        labels.insert(0, pd.Timestamp(start_time).date())
        ticks.insert(0, 0)

        ax.set_xticks(ticks)
        ax.set_xticklabels(labels, rotation=45)

    elif diff < sec_per_half_year:
         # generate label every two days
        ticks, nr_tick2first_tick = get_stuff_done(tick_step=sec_per_day*2, int_range=diff,
                                                   start_time=start_time)

        dates = pd.date_range(start_time + pd.Timedelta('2D'), periods=len(ticks), freq='2D')
        labels = [date.date() for date in dates]

        ax.set_xticks(ticks)
        ax.set_xticklabels(labels, rotation=45)
    else:
         # generate label every week
        ticks, nr_tick2first_tick = get_stuff_done(tick_step=sec_per_day*7, int_range=diff,
                                                   start_time=start_time)

        dates = pd.date_range(start_time, periods=len(ticks), freq='7D')
        labels = [date.date() for date in dates]

        ax.set_xticks(ticks)
        ax.set_xticklabels(labels, rotation=45)


def add_top_axis_time_format(axis, top_label=None):
    ax_top = axis.secondary_xaxis('top', functions=(lambda x: x, lambda x: x))
    ax_top.xaxis.set_major_formatter(
        ticker.FuncFormatter(func_formatter_seconds2time))
    if top_label is not None:
        ax_top.set_xlabel(top_label)


def plot_hist(ax, bins, counts, color='k'):
    """ plot a histogram given bins and counts for the bins
    from https://gist.github.com/hitvoice/47c63728754713d0e56eb8366bfafe56
    """
    assert len(bins) == len(counts) + 1
    centroids = (bins[1:] + bins[:-1]) / 2  # values in the middle between each bin

    # plot for every centroid one occurrence  that is multiplied with the weight -> real height
    counts_, bins_, _ = ax.hist(centroids, bins=len(counts), weights=counts, range=(min(bins), max(bins)), color=color)
    assert np.allclose(bins_, bins)
    assert np.allclose(counts_, counts)

def plot_cc(ccg, bins, title, y_label=None, x_label=None, axis='off', figsize=None, use_dask=False):
    """
    Plots a cross correlogram of histograms
    """

    I = ccg.shape[0]    # nr devices
    J = ccg.shape[1]

    # plot
    colors = plt.cm.jet(np.linspace(0, 1, I+1))
    def infer_figsize(K):
        if K < 10:
            return (6, 6)
        elif K < 20:
            return (10, 10)

    if figsize is None:
        figsize = infer_figsize(ccg.shape[1])
    fig = plt.figure(2, figsize=figsize)
    bg = 0.7*np.ones(3)
    fig.suptitle(title, fontsize=14)

    import matplotlib.gridspec as gridspec
    gs = gridspec.GridSpec(I, J)
    gs.update(wspace=0.05, hspace=0.05) # set the spacing between axes.

    def set_x_ticks(ax , n_devices):
        if bins[-1] < 60:
            bound = str(int(bins[-1])) + 's'
            ax.set_xticks([bins[0], 0, bins[-1]])
            ax.set_xticklabels(['-' + bound, '0', bound], rotation='vertical')
        elif bins[-1] < 180:
            bound = '{:.1f}'.format(bins[-1]/60) + 'm'
            ax.set_xticks([bins[0], 0, bins[-1]])
            ax.set_xticklabels(['-' + bound, '0', bound], rotation='vertical')
        elif bins[-1] < 600:
            ax.set_xticks([bins[0], 0, bins[-1]])
            bound = '{:.1f}'.format(bins[-1]/60) + 'm'
            ax.set_xticks([bins[0], -300, 300, bins[-1]])
            ax.set_xticklabels(['-' + bound, '-5m', '5m', bound], rotation='vertical')
        elif bins[-1] < 3600:
            ax.set_xticks([bins[0], 0, bins[-1]])
            bound = '{:.1f}'.format(bins[-1]/60) + 'm'
            ax.set_xticks([bins[0], -60, 60, bins[-1]])
            ax.set_xticklabels(['-' + bound, '-1m', '1m', bound], rotation='vertical')


    if not use_dask:
        for ix in range(I):
            for jx in range(J):
                ax = plt.subplot(gs[J*ix+jx], facecolor=bg)
                #ax = plt.subplot(K,K,K*ix+jx+1, facecolor=bg)

                counts = ccg[ix, jx, :]
                # remove middle count because it is weird for a histogram to have.. TODO check in correlogram method
                counts = np.concatenate([counts[:int(len(counts)/2)],
                                    counts[int(len(counts)/2)+1:]])
                if ix == jx:
                    plot_hist(ax, bins, counts, color=colors[ix, :])
                else:
                    plot_hist(ax, bins, counts, color='k')

                ax.set_xlim(1.2 * bins[[0, -1]])
                ylim = np.array(list(ax.get_ylim()))
                ax.set_ylim(np.array([0, 1.2]) * ylim)
                ax.set_yticks([])

                if ix == 0 and x_label is not None:
                    sec_ax = ax.secondary_xaxis('top')
                    sec_ax.set_xticks([])
                    sec_ax.set_xlabel(x_label[jx], rotation=45)

                if ix != I-1 and ix != 0:
                    ax.set_xticks([])

                if ix == I-1:
                    set_x_ticks(ax, I)

                if jx == 0:
                    ax.set_yticks([])
                    if y_label is not None:
                        ax.set_ylabel(y_label[ix], rotation='horizontal', ha='right')

                if ix != jx:
                    ax.plot(0, 0, '*', c=colors[jx, :])
        #plt.tight_layout()
        fig.show()
        return fig
    else:
        def func(ax, ix, jx):
            counts = ccg[ix, jx, :]
            # remove middle count because it is weird for a histogram to have.. TODO check in correlogram method
            counts = np.concatenate([counts[:int(len(counts)/2)],
                                counts[int(len(counts)/2)+1:]])
            if ix == jx:
                plot_hist(ax, bins, counts, color=colors[ix, :])
            else:
                plot_hist(ax, bins, counts, color='k')

            ax.set_xlim(1.2 * bins[[0, -1]])
            ylim = np.array(list(ax.get_ylim()))
            ax.set_ylim(np.array([0, 1.2]) * ylim)
            ax.set_yticks([])

            if ix == 0 and x_label is not None:
                sec_ax = ax.secondary_xaxis('top')
                sec_ax.set_xticks([])
                sec_ax.set_xlabel(x_label[jx], rotation=45)

            if ix != I-1 and ix != 0:
                ax.set_xticks([])

            if ix == I-1:
                set_x_ticks(ax, I)

            if jx == 0:
                ax.set_yticks([])
                if y_label is not None:
                    ax.set_ylabel(y_label[ix], rotation='horizontal', ha='right')

            if ix != jx:
                ax.plot(0, 0, '*', c=colors[jx, :])
        import dask

        res = []
        for ix in range(I):
            for jx in range(J):
                ax = plt.subplot(gs[J*ix+jx], facecolor=bg)
                #ax = plt.subplot(K,K,K*ix+jx+1, facecolor=bg)
                res.append(dask.delayed(func)(ax, ix, jx))
        res = dask.compute(*res)
        fig.show()
        return fig

def save_fig(func):
    """ Decorator that saves a figure to a filepath rather then returning it
        if the keyword file_path is given.
    """
    @functools.wraps(func)
    def wrapper_save_fig(*args, **kwargs):
        fig = func(*args, **kwargs)
        try:
            fp = kwargs['file_path']
            savefig(fig, fp)
            return None
        except:
            return fig

    return wrapper_save_fig


def _format_dev_and_state(word):
    return ''.join(' - ' if c == ':' else c for c in word)

def _only_state(word):
    return word.split(':')[1]

def _only_dev(word):
    return word.split(':')[0]




def plot_grid(fig, ax, start_time, end_time):
    """ In a plot that spans a time. Plot depending on the date range a grid.
        For example,

    """
    new_ax = fig.add_axes(ax.get_position(), frameon=False)
    new_ax.plot([start_time, end_time], [0, 0], alpha=0.0)
    new_ax.set_yticks([])

    from matplotlib.dates import SecondLocator, MinuteLocator, HourLocator, DayLocator
    range = end_time - start_time
    minor_locator = None

    # lesser than one day use HourLocator
    # greater than 3 days use day locator
    if range <= pd.Timedelta('5H'):
        major_locator = HourLocator()
        minor_locator = MinuteLocator(byminute=np.arange(0, 60, 10))
    elif range <= pd.Timedelta('1D'):
        # major locator at every 2 hours and minor 2 hours
        major_locator = HourLocator(byhour=np.arange(0, 24, 2))
    elif range <= pd.Timedelta('7D'):
        # major every day and minor every half day
        major_locator = DayLocator()
        minor_locator = HourLocator(byhour=np.arange(0, 24, 6))
    elif range <= pd.Timedelta('2W'):
        # major every day and minor every half day
        major_locator = DayLocator()
        minor_locator = HourLocator(byhour=np.arange(0, 24, 12))
    else:
        major_locator = DayLocator()

    new_ax.xaxis.set_major_locator(major_locator)
    new_ax.grid(b=True, which='major', axis='x', linestyle='-')
    if minor_locator is not None:
        new_ax.xaxis.set_minor_locator(minor_locator)
        new_ax.grid(b=True, which='minor', axis='x', linestyle='--')
    new_ax.set_xticklabels([])




# Colors for plots
PRIMARY_COLOR = '#2c3e50'
SECONDARY_COLOR = '#e34d30'
CM_DIV_MAP = 'RdBu'     # RdGy
CM_SEQ_MAP = 'viridis'  # BrwnYI, BurgYI, Burg

# String representations that are used to format quantities
# for plotly graphs
STRFTIME_DATE = '%d.%m.%Y %H:%M:%S.%s'
STRFTIME_HOUR = '%H:%M:%S'
STRFTIME_DELTA = ''
STRFTIME_PRECISE = '%d.%m.%Y %H:%M:%S.%f' # Microsecond accuracy

# Name for the other activity; used to fill in gaps between activities
OTHER = 'other'

# Dataset columns

# df_activities columns
START_TIME = 'start_time'
END_TIME = 'end_time'
ACTIVITY = 'activity'
# df_devices columns
TIME = 'time'
DEVICE = 'device'
VALUE = 'value'
NAME = 'name'

# Device data types
CAT = 'categorical'
NUM = 'numerical'
BOOL = 'boolean'

# Device encoding
ENC_RAW = 'raw'
ENC_LF = 'last_fired'
ENC_CP = 'changepoint'
REPS = [ENC_RAW, ENC_LF, ENC_CP]


# Activity assistant
AREA = 'area'


DATA_DUMP_NAME = 'data.joblib'
DATA_HOME_FOLDER_NAME = 'pyadlml_data_home'


# The keys of dictionary returned by fetch methods
DATA_DCT_KEY_DEVS = 'devices'
DATA_DCT_KEY_ACTS = 'activities'

# Dataset identifiers
DATASET_STRINGS = [
    'casas_aruba',
    'casas_kyoto_2010',
    'casas_milan',
    'casas_cairo',
    'casas_tulum',
    'amsterdam',
    'mitlab_1',
    'mitlab_2',
    'aras',
    'kasteren_A',
    'kasteren_B',
    'kasteren_C',
    'tuebingen_2019',
    'uci_ordonezA',
    'uci_ordonezB',
    'act_assist'
]

import pandas as pd
ts2str = lambda ts: ts.strftime(STRFTIME_PRECISE)
str2ts = lambda s: pd.to_datetime(s, dayfirst=True)


from pyadlml.dataset._datasets.amsterdam import fetch_amsterdam
from pyadlml.dataset._datasets.aras import fetch_aras
from pyadlml.dataset._datasets.mitlab import fetch_mitlab
from pyadlml.dataset._datasets.tuebingen_2019 import fetch_tuebingen_2019
from pyadlml.dataset._datasets.uci_adl_binary import fetch_uci_adl_binary
from pyadlml.dataset._datasets.kasteren_2010 import fetch_kasteren_2010
from pyadlml.dataset._datasets.casas_houses import fetch_casas

from pyadlml.dataset._datasets.homeassistant import (
    load_homeassistant, load_homeassistant_devices
)

from pyadlml.dataset._datasets.activity_assistant import load as load_act_assist
from pyadlml.dataset import io

# TODO refactor, importing stuff from pyadlml.dataset must not require 
#                torch as a dependency to be installed. 
#from pyadlml.dataset.misc import TorchDataset



import numpy as np
import pandas as pd

from pyadlml.constants import DEVICE, TIME, VALUE, CAT, NUM, BOOL
from pyadlml.dataset._core.devices import _create_devices, most_prominent_categorical_values
from pyadlml.dataset.util import infer_dtypes

ST_FFILL = 'ffill'
ST_INT_COV = 'interval_coverage'

def create_raw(df_dev, dataset_info, dev_pre_values={}):
    """

    Parameters
    ----------
    df_dev : pd.DataFrame

    dataset_info : dict
        first key: devices (DEVICE)
        per dev key: most likely value ('ml_state')
        per dev key: datatype ('dtype')
    dev_pre_values : dict
        a dictionary a mapping from device to values. This mapping should be
        used for values where the preceeding value is not known.

    Returns
    -------
        return df:
        | time  | dev_1 | ....  | dev_n |
        --------------------------------
        | ts1   |   1   | ....  | open |
    """
    df_dev = df_dev.copy()

    df = df_dev.pivot(index=TIME, columns=DEVICE, values=VALUE)
    df = df.reset_index()

    # get all learned devices by data type
    dev_cat = [dev for dev in dataset_info.keys() if dataset_info[dev]['dtype'] == CAT]
    dev_bool = [dev for dev in dataset_info.keys() if dataset_info[dev]['dtype'] == BOOL]
    dev_num = [dev for dev in dataset_info.keys() if dataset_info[dev]['dtype'] == NUM]

    # filter for devices that appear in given dataset
    devs = set(df_dev[DEVICE].unique())
    dev_cat = list(set(dev_cat).intersection(devs))
    dev_bool = list(set(dev_bool).intersection(devs))
    dev_num = list(set(dev_num).intersection(devs))

    # set the first element for each boolean device to the opposite value of the
    # first occurrence
    for dev in dev_bool:
        fvi = df[dev].first_valid_index()
        if fvi != 0:
            value = df[dev].iloc[fvi]
            df.loc[0, dev] = not value

    # set the first element of each categorical device to the most likely value
    for dev in dev_cat:
        if dev_pre_values:
            new_val = dev_pre_values[dev]
        else:
            new_val = dataset_info[dev]['ml_state']
        df.loc[0, dev] = new_val

    # set the first element of numerical values to the given value if dev_pre_values
    # dict is given
    for dev in dev_num:
        if dev_pre_values:
            new_val = dev_pre_values[dev]
            df.loc[0, dev] = new_val

    # fill from start to end NaNs with the preceding correct value
    df_cat_bool = df[list(dev_bool) + list(dev_cat)].ffill()

    # join all dataframes
    df = pd.concat([df[TIME], df[dev_num], df_cat_bool], axis=1)

    # for all devices that are present in the info but not in the current dataframe infer value
    for dev in (set(dataset_info.keys()) - set(df.columns)):
        if dev_pre_values:
            df[dev] = dev_pre_values[dev]
        else:
            df[dev] = dataset_info[dev]['ml_state']
    return df


def resample_raw(df_raw, df_dev, dt, most_likely_values=None, n_jobs=1):
    """
    Resamples a raw representation
    """

    # get dtypes in order for choosing different collision behavior
    dev_dtypes = infer_dtypes(df_dev)
    if most_likely_values is None:
        from pyadlml.dataset.devices import get_most_likely_value
        most_likely_values = get_most_likely_value(df_dev)
        most_likely_values = pd.DataFrame(most_likely_values).T
    variant = 'a'
    if variant == 'a':
        return resample_pandas(df_raw, dt, df_dev, dev_dtypes, most_likely_values)
    elif variant == 'c':
        return resample_pandas_and_pandarell_applymap(df_raw, dt, df_dev, dev_dtypes, most_likely_values)
    elif variant == 'd':
        return resample_dask_pandarell(df_raw, dt, dev_dtypes, df_dev, most_likely_values)
    elif variant == 'b':
        return resample_pure_dask(df_raw, dt, dev_dtypes, df_dev, most_likely_values)


"""
FROM HERE ON ARE THE DIFFERENT resample techniques. As of now pandas and pandarell applymap 
is the fastest.
"""


def parallel_resample_apply(resampler, func):
    """
    https://stackoverflow.com/questions/26187759/parallelize-apply-after-pandas-groupby
    # TODO write manual parallelism
    """
    import pandas as pd
    from joblib import Parallel, delayed
    import multiprocessing
    jobs = []
    for name, group in resampler:
        jobs.append(delayed(func)(group))
    ret_lst = Parallel(n_jobs=multiprocessing.cpu_count())(jobs)
    #retLst = Parallel(n_jobs=multiprocessing.cpu_count())(delayed(func)(group) for name, group in grouped)
    return pd.concat(retLst)

def resample_pandas(df_raw, t_res, df_dev, dev_dtypes, most_likely_values):
        df_raw = df_raw.set_index(TIME)
        resampler = df_raw.resample(t_res, kind='timestamp')

        # first do a forward fill to correctly represent intervals where no observation falls into
        # ffill takes a devices last known value and assigns it to the timeslice
        raw_ff = resampler.ffill()

        # then for intervals where multiple sensors trigger choose the most prevalent for categorical and
        # boolean values. For numerical drop the most unlikely one (determined on difference to mean)
        raw_int = resampler.apply(_assign_timeslices,
                                  t_res=t_res,
                                  dev=df_dev,
                                  dev_dtypes=dev_dtypes,
                                  most_likely_values=most_likely_values
                                  )
        # combine both by filling gaps with the forward fills
        raw = raw_int.where(~raw_int.isnull(), raw_ff)

        raw = raw.reset_index()
        return raw


def resample_pandas_and_pandarell_applymap(df_raw, t_res, df_dev, dev_dtypes, most_likely_value):
    """ use pandas with and pandarallel applymap """
    from pandarallel import pandarallel
    pandarallel.initialize(verbose=1)

    df_raw = df_raw.set_index(TIME)
    raw_ff = df_raw.resample(t_res, kind='timestamp').ffill()

    # first make resampling and safe lists of choices inside each bin
    raw_int = df_raw.resample(t_res).apply(lambda x: [x.name, x.tolist(), x.index.tolist()])

    def _assign_timeslicesC(lst: list) -> pd.Series:
        """
        cc. kasteren

        Parameters
        ----------
        series: pd.Series
            the datapoints that are to be assigned for a timeslice for one column/device
        t_res: the resolution of the intervals
            e.g 30s ->  09:38:30, 09:39:00
        devs: pd.DataFrame
            used to identify when values changes

        """
        name = lst[0]
        values = lst[1]
        timestamps = lst[2]
        series = pd.Series(values, index=timestamps)
        series.name = lst[0]

        # return nan if no element matchs, happens till first occurence of data
        series = series.dropna()
        if series.empty:
            return np.nan

        # if one element matches the interval slot, assign the one
        elif series.size == 1:
            return series

        # if there are multiple elements falling into the same timeslice
        else:
            if series.name in dev_dtypes['boolean']:
                tmp = resolve_collision_boolean(series, t_res, df_dev)

            elif series.name in dev_dtypes['categorical']:
                tmp = resolve_collision_categorical(series, t_res, df_dev)

            elif series.name in dev_dtypes['numerical']:
                tmp = resolve_collision_numerical(series, t_res, df_dev, most_likely_value)
            else:
                raise ValueError("The column/device didn't match either boolean, categorical or numerical type.")
            return tmp

    # on each field apply the collision resolvment
    #raw_int = raw_int.applymap(_assign_timeslicesC)
    #raw_int = raw_int.parallel_applymap(_assign_timeslicesC)

    # combine both by filling gaps with the forward fills
    raw = raw_int.where(~raw_int.isnull(), raw_ff)

    raw = raw.reset_index()
    return raw

def resample_pure_dask(df_raw, t_res, dev_dtypes, df_dev, most_likely_values):
        import dask.dataframe as dd
        dd_raw = dd.from_pandas(df_raw, 4)

        # TODO problem dask resampler doesn't implement function ffill
        # solution 1
        resampler_ffill = dd_raw.groupby(pd.Grouper(key=TIME, freq=t_res, origin='start'))
        def func(x):
            """ returns nan if list is empty and """
            x = x.dropna()
            if x.empty:
                return np.nan
            else:
                return x.iloc[-1]

        raw_ff = resampler_ffill.apply(func)
        df = raw_ff.compute()
        # solution 2: use rolling in combination with apply and a lambda
        # function that mimics ffill to achieve parallelnes

        resampler = dd_raw.resample(t_res)
        raw_int = resampler.apply(_assign_timeslices, t_res=t_res, dev=df_dev,
                              dev_dtypes=dev_dtypes, most_likely_values=most_likely_values)
        raw = raw_int.where(~raw_int.isnull(), raw_ff)
        raw = raw.reset_index().compute()
        return raw


def resample_dask_pandarell(df_raw, t_res, dev_dtypes, df_dev, most_likely_values):
        import dask.dataframe as dd
        n_jobs = 8
        df_raw = df_raw.set_index(TIME)
        raw_ff = df_raw.resample(t_res, kind='timestamp').ffill()

        # first make resampling and safe lists of choices inside each bin
        dd_raw_int = dd.from_pandas(df_raw, npartitions=n_jobs)
        raw_count = dd_raw_int.resample(t_res).count().compute()
        row_mask = (raw_count > 0).any(axis=1)
        # first make resampling and safe lists of choices inside each bin
        #raw_int = df_raw.resample(t_res).apply(lambda x: [x.name, x.tolist(), x.index.tolist()])
        raw_int = parallel_resample_apply(df_raw.resample(t_res), lambda x: [x.name, x.tolist(), x.index.tolist()])
        raw_int_subset = raw_int.loc[row_mask]

        def _assign_timeslicesC(lst: list) -> pd.Series:
            name = lst[0]
            values = lst[1]
            timestamps = lst[2]
            series = pd.Series(values, index=timestamps)
            series.name = name
            # return nan if no element matchs, happens till first occurence of data
            series = series.dropna()
            if series.empty:
                return np.nan

            # if one element matches the interval slot, assign the one
            elif series.size == 1:
                return series

            # if there are multiple elements falling into the same timeslice
            else:
                if series.name in dev_dtypes['boolean']:
                    tmp = resolve_collision_boolean(series, t_res, df_dev)

                elif series.name in dev_dtypes['categorical']:
                    tmp = resolve_collision_categorical(series, t_res, df_dev)

                elif series.name in dev_dtypes['numerical']:
                    tmp = resolve_collision_numerical(series, t_res, df_dev, most_likely_values)
                else:
                    raise ValueError("The column/device didn't match either boolean, categorical or numerical type.")
                return tmp

        from pandarallel import pandarallel
        pandarallel.initialize(verbose=2)
        # on each field apply the collision resolvment
        raw_int_subset = raw_int_subset.parallel_apply(_assign_timeslicesC)
        raw_int = pd.concat(raw_int_subset, raw_int[~row_mask]).sort_values(by=TIME)

        # combine both by filling gaps with the forward fills
        raw = raw_int.where(~raw_int.isnull(), raw_ff)

        raw = raw.reset_index()
        return raw

def _assign_timeslices(series: pd.Series, t_res, dev, dev_dtypes, most_likely_values) -> pd.Series:
    """
    cc. kasteren
    
    Parameters
    ----------
    series: pd.Series
        the datapoints that are to be assigned for a timeslice for one column/device
    t_res: the resolution of the intervals
        e.g 30s ->  09:38:30, 09:39:00
    devs: pd.DataFrame
        used to identify when values changes

    """
    # return nan if no element matchs, happens till first occurence of data
    series = series.dropna()
    if series.empty:
        return np.nan

    # if one element matches the interval slot, assign the one
    elif series.size == 1:
        return series

    # if there are multiple elements falling into the same timeslice
    else:
        if series.name in dev_dtypes['boolean']:
            tmp = resolve_collision_boolean(series, t_res, dev)

        elif series.name in dev_dtypes['categorical']:
            tmp = resolve_collision_categorical(series, t_res, dev)

        elif series.name in dev_dtypes['numerical']:
            tmp = resolve_collision_numerical(series, t_res, dev, most_likely_values)
        else:
            raise ValueError("The column/device didn't match either boolean, categorical or numerical type.")
        return tmp


def resolve_collision_numerical(series: pd.Series, t_res, dev, most_likely_values: pd.DataFrame):
    """
    The value that is nearer to the mean value is treated as more likely

    Discussion: maybe the value that is further away from the mean carries more information and
    should therefore be kept ???
    """
    dev_mean = most_likely_values.at[series.name, 'ml_state']
    # get elements position with minimal distance to most likely value (mean/median)
    min_idx = abs(series - dev_mean).values.argmin()
    val = series.iloc[min_idx]
    return val


def resolve_collision_categorical(series, t_res, df_dev):
    """
    choose the element with the maximal overlap
    """
    dev_name = series.name

    # create a category timedelta score dictionary
    df_subdev = df_dev[df_dev[DEVICE] == dev_name]
    cats = list(df_subdev[VALUE].unique())
    tds = [pd.Timedelta(seconds=0) for i in range(len(cats))]
    categories = dict(zip(cats,tds))

    # deduce the beginning of the interval and the timeslice
    start_time = series.index[0].floor(t_res)
    end_time = start_time + pd.Timedelta(t_res)

    # get preceding categorical value
    prec_values = df_subdev[df_subdev[TIME] < start_time]
    if prec_values.empty:
        # if this is the first slice there are no preceding values so take the first of the series
        prae_val = series[0]
    else:
        # take the last known category as
        prae_val = prec_values.iat[-1, 2]
    prae_ts = start_time

    # compute cumulative states in each sensor
    for entry in series.iteritems():
        cur_ts = entry[0]
        categories[prae_val] += cur_ts - prae_ts
        prae_val = entry[1]
        prae_ts = cur_ts

    # add time to end of timeslice to the last known category
    categories[prae_val] += end_time - prae_ts

    # select category with most overlap
    max_cat = max(categories, key=lambda key: categories[key])
    return max_cat

def resolve_collision_boolean(series, t_res, dev):
    """
    choose the element with the maximal overlap
    """
    cum_time_0 = pd.Timedelta(seconds=0)
    cum_time_1 = pd.Timedelta(seconds=0)

    # deduce the beginning of the interval and the interval
    start_time = series.index[0].floor(t_res)
    end_time = start_time + pd.Timedelta(t_res)

    # deduce value at start of the timeslice
    df_subdev = dev[dev[DEVICE] == series.name]
    prec_values = df_subdev[df_subdev[TIME] < start_time]
    if prec_values.empty:
        # if this is the first slice there are no preceding values so take the first of the series
        prae_val = series[0]
    else:
        # take the last known category as
        prae_val = prec_values.iat[-1, 2]

    prae_time = start_time
    # compute cumulative states in each sensor
    for entry in series.iteritems():
        ts = entry[0]
        cur_val = entry[1]
        if prae_val:
            cum_time_1 += ts-prae_time
        else:
            cum_time_0 += ts-prae_time
        prae_val = cur_val
        prae_time = ts

    # compute cum time to remaining end of interval
    if prae_val:
        cum_time_1 += end_time-prae_time
    else:
        cum_time_0 += end_time-prae_time

    assert cum_time_0 + cum_time_1 == pd.Timedelta(t_res)

    # return the state that was more present in the interval
    return int(cum_time_0 < cum_time_1)


"""
Full definition of a GPT Language Model, all of it in this single file.
References:
1) the official GPT-2 TensorFlow implementation released by OpenAI:
https://github.com/openai/gpt-2/blob/master/src/model.py
2) huggingface/transformers PyTorch implementation:
https://github.com/huggingface/transformers/blob/main/src/transformers/models/gpt2/modeling_gpt2.py


Stolen from karpathy

batch_size = 64 # how many independent sequences will we process in parallel?
block_size = 256 # what is the maximum context length for predictions?
max_iters = 5000
eval_interval = 500
learning_rate = 3e-4
device = 'cuda' if torch.cuda.is_available() else 'cpu'
eval_iters = 200
n_embd = 384
n_head = 6
n_layer = 6
dropout = 0.2


"""

import math
import inspect
from dataclasses import dataclass
import torch
import torch.nn as nn
from torch.nn import functional as F

# @torch.jit.script # good to enable when not using torch.compile, disable when using (our default)
def new_gelu(x):
    """
    Implementation of the GELU activation function currently in Google BERT repo (identical to OpenAI GPT).
    Reference: Gaussian Error Linear Units (GELU) paper: https://arxiv.org/abs/1606.08415
    """
    return 0.5 * x * (1.0 + torch.tanh(math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0))))

class LayerNorm(nn.Module):
    """ LayerNorm but with an optional bias. PyTorch doesn't support simply bias=False """

    def __init__(self, ndim, bias):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(ndim))
        self.bias = nn.Parameter(torch.zeros(ndim)) if bias else None

    def forward(self, input):
        return F.layer_norm(input, self.weight.shape, self.weight, self.bias, 1e-5)

class CausalSelfAttention(nn.Module):

    def __init__(self, n_embd, n_head, dropout, bias, block_size):
        super().__init__()
        assert n_embd % n_head == 0
        # key, query, value projections for all heads, but in a batch
        self.c_attn = nn.Linear(n_embd, 3 * n_embd, bias=bias)
        # output projection
        self.c_proj = nn.Linear(n_embd, n_embd, bias=bias)
        # regularization
        self.attn_dropout = nn.Dropout(dropout)
        self.resid_dropout = nn.Dropout(dropout)
        self.n_head = n_head
        self.n_embd = n_embd
        self.dropout = dropout
        # flash attention make GPU go brrrrr but support is only in PyTorch nightly and still a bit scary
        self.flash = hasattr(torch.nn.functional, 'scaled_dot_product_attention') and self.dropout == 0.0
        if not self.flash:
            print("WARNING: using slow attention. Flash Attention atm needs PyTorch nightly and dropout=0.0")
            # causal mask to ensure that attention is only applied to the left in the input sequence
            self.register_buffer("bias", torch.tril(torch.ones(block_size, block_size))
                                        .view(1, 1, block_size, block_size))

    def forward(self, x):
        B, T, E = x.size() # batch size, sequence length, embedding dimensionality (n_embd)

        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        # q -> (B, T, E)
        q, k ,v  = self.c_attn(x).split(self.n_embd, dim=2)
        k = k.view(B, T, self.n_head, E // self.n_head).transpose(1, 2) # (B, nh, T, hs)
        q = q.view(B, T, self.n_head, E // self.n_head).transpose(1, 2) # (B, nh, T, hs)
        v = v.view(B, T, self.n_head, E // self.n_head).transpose(1, 2) # (B, nh, T, hs)

        # causal self-attention; Self-attend: (B, nh, T, hs) x (B, nh, hs, T) -> (B, nh, T, T)
        if self.flash:
            # efficient attention using Flash Attention CUDA kernels
            y = torch.nn.functional.scaled_dot_product_attention(q, k, v, attn_mask=None, dropout_p=self.dropout, is_causal=True)
        else:
            # manual implementation of attention
            att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
            att = att.masked_fill(self.bias[:,:,:T,:T] == 0, float('-inf'))
            att = F.softmax(att, dim=-1)
            att = self.attn_dropout(att)
            y = att @ v # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)
        y = y.transpose(1, 2).contiguous().view(B, T, E) # re-assemble all head outputs side by side

        # output projection
        y = self.resid_dropout(self.c_proj(y))
        return y

class MLP(nn.Module):

    def __init__(self, n_embd, bias, dropout):
        super().__init__()
        self.c_fc    = nn.Linear(n_embd, 4 * n_embd, bias=bias)
        self.c_proj  = nn.Linear(4 * n_embd, n_embd, bias=bias)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        x = self.c_fc(x)
        x = new_gelu(x)
        x = self.c_proj(x)
        x = self.dropout(x)
        return x

class DecoderBlock(nn.Module):

    def __init__(self, n_embd, bias, mlp_kwargs, csa_kwargs):
        super().__init__()
        self.ln_1 = LayerNorm(n_embd, bias=bias)
        self.attn = CausalSelfAttention(
            n_embd=n_embd,
            n_head=csa_kwargs['n_head'],
            dropout=csa_kwargs['dropout'],
            bias=bias,
            block_size=csa_kwargs['block_size']
        )
        self.ln_2 = LayerNorm(n_embd, bias=bias)
        self.mlp = MLP(
            n_embd=n_embd,
            bias=bias, 
            dropout=mlp_kwargs['dropout']
        )

    def forward(self, x: torch.tensor):
        """
        Parameters
        ----------
        x : tensor of shape (B, T, E)

        Returns
        -------
        tensor of shape (TODO )
        """
        x = x + self.attn(self.ln_1(x)) # (B, T, E) -> (B, T, E)
        x = x + self.mlp(self.ln_2(x))  # (B, T, E) -> (B, T, E)
        return x

#@dataclass
#class GPTConfig:
#    block_size: int = 1024
#    vocab_size: int = 50304 # GPT-2 vocab_size of 50257, padded up to nearest multiple of 64 for efficiency
#    n_layer: int = 12
#    n_head: int = 12
#    n_embd: int = 768
#    dropout: float = 0.0
#    bias: bool = True # True: bias in Linears and LayerNorms, like GPT-2. False: a bit better and faster

class VanillaTransformer(nn.Module):

    def __init__(self, n_embd: int, block_size: int, n_features: int, bias: bool, n_layer: int, n_head: int, n_classes: int, dropout: float):
        super().__init__()
        assert n_features is not None
        assert block_size is not None
        self.block_size = block_size

        self.transformer = nn.ModuleDict(dict(
            wte = nn.Embedding(n_features, n_embd),
            wpe = nn.Embedding(block_size, n_embd),
            drop = nn.Dropout(dropout),
            h = nn.ModuleList([DecoderBlock(
                n_embd, bias, mlp_kwargs=dict(
                    dropout=dropout
                ), 
                csa_kwargs=dict(
                    n_head=n_head,
                    dropout=dropout,
                    block_size=block_size,
                )
            ) for _ in range(n_layer)]),
            ln_f = LayerNorm(n_embd, bias=bias),
        ))
        self.lm_head = nn.Linear(n_embd, n_classes, bias=False)

        # Classes/activities that are predicted are not equal to 
        # input
        #self.transformer.wte.weight = self.lm_head.weight # https://paperswithcode.com/method/weight-tying

        # init all weights
        self.apply(self._init_weights)

        # apply special scaled init to the residual projections, per GPT-2 paper
        for pn, p in self.named_parameters():
            if pn.endswith('c_proj.weight'):
                torch.nn.init.normal_(p, mean=0.0, std=0.02/math.sqrt(2 * n_layer))

        # report number of parameters
        print("number of parameters: %.2fM" % (self.get_num_params()/1e6,))

    def get_num_params(self, non_embedding=True):
        """
        Return the number of parameters in the model.
        For non-embedding count (default), the position embeddings get subtracted.
        The token embeddings would too, except due to the parameter sharing these
        params are actually used as weights in the final layer, so we include them.
        """
        n_params = sum(p.numel() for p in self.parameters())
        if non_embedding:
            n_params -= self.transformer.wpe.weight.numel()
        return n_params

    def _init_weights(self, module):
        if isinstance(module, nn.Linear):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            if module.bias is not None:
                torch.nn.init.zeros_(module.bias)
        elif isinstance(module, nn.Embedding):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)

    def compute_loss(self, logits, y_true, weights):
        return F.cross_entropy(logits, y_true, weight=weights) 
                                    

    def forward(self, idx):
        device = idx.device
        B, T = idx.size()
        assert T <= self.block_size, f"Cannot forward sequence of length {T}, block size is only {self.block_size}"
        pos = torch.arange(0, T, dtype=torch.long, device=device).unsqueeze(0) # shape (1, t)

        # forward the GPT model itself
        tok_emb = self.transformer.wte(idx) # token embeddings of shape (B, T, E)
        pos_emb = self.transformer.wpe(pos) # position embeddings of shape (B, T, E)
        # TODO pos_emb is (1, T, E) 
        x = self.transformer.drop(tok_emb + pos_emb)

        # Iterate over all blocks
        for block in self.transformer.h:
            x = block(x)
        x = self.transformer.ln_f(x)

        # if we are given some desired targets also calculate the loss
        logits = self.lm_head(x)
        if False:
            # TODO refactor this some time
            # inference-time mini-optimization: only forward the lm_head on the very last position
            # (B, T, E) -> (B, 1, E)
            logits = self.lm_head(x[:, [-1], :]) # note: using list [-1] to preserve the time dim

        return logits

    def configure_optimizers(self, weight_decay, learning_rate, betas, device_type):
        """
        This long function is unfortunately doing something very simple and is being very defensive:
        We are separating out all parameters of the model into two buckets: those that will experience
        weight decay for regularization and those that won't (biases, and layernorm/embedding weights).
        We are then returning the PyTorch optimizer object.
        """

        # separate out all parameters to those that will and won't experience regularizing weight decay
        decay = set()
        no_decay = set()
        whitelist_weight_modules = (torch.nn.Linear, )
        blacklist_weight_modules = (torch.nn.LayerNorm, LayerNorm, torch.nn.Embedding)
        for mn, m in self.named_modules():
            for pn, p in m.named_parameters():
                fpn = '%s.%s' % (mn, pn) if mn else pn # full param name
                # random note: because named_modules and named_parameters are recursive
                # we will see the same tensors p many many times. but doing it this way
                # allows us to know which parent module any tensor p belongs to...
                if pn.endswith('bias'):
                    # all biases will not be decayed
                    no_decay.add(fpn)
                elif pn.endswith('weight') and isinstance(m, whitelist_weight_modules):
                    # weights of whitelist modules will be weight decayed
                    decay.add(fpn)
                elif pn.endswith('weight') and isinstance(m, blacklist_weight_modules):
                    # weights of blacklist modules will NOT be weight decayed
                    no_decay.add(fpn)

        # subtle: 'transformer.wte.weight' and 'lm_head.weight' are tied, so they
        # will appear in the no_decay and decay sets respectively after the above.
        # In addition, because named_parameters() doesn't return duplicates, it
        # will only return the first occurence, key'd by 'transformer.wte.weight', below.
        # so let's manually remove 'lm_head.weight' from decay set. This will include
        # this tensor into optimization via transformer.wte.weight only, and not decayed.
        decay.remove('lm_head.weight')

        # validate that we considered every parameter
        param_dict = {pn: p for pn, p in self.named_parameters()}
        inter_params = decay & no_decay
        union_params = decay | no_decay
        assert len(inter_params) == 0, "parameters %s made it into both decay/no_decay sets!" % (str(inter_params), )
        assert len(param_dict.keys() - union_params) == 0, "parameters %s were not separated into either decay/no_decay set!" \
                                                    % (str(param_dict.keys() - union_params), )

        # create the pytorch optimizer object
        optim_groups = [
            {"params": [param_dict[pn] for pn in sorted(list(decay))], "weight_decay": weight_decay},
            {"params": [param_dict[pn] for pn in sorted(list(no_decay))], "weight_decay": 0.0},
        ]
        # new PyTorch nightly has a new 'fused' option for AdamW that is much faster
        use_fused = (device_type == 'cuda') and ('fused' in inspect.signature(torch.optim.AdamW).parameters)
        print(f"using fused AdamW: {use_fused}")
        extra_args = dict(fused=True) if use_fused else dict()
        optimizer = torch.optim.AdamW(optim_groups, lr=learning_rate, betas=betas, **extra_args)

        return optimizer




from .downloader import MegaDownloader
from .remote import DataFetcher
from .local import get_data_home, set_data_home, clear_data_home


__all__ = [
    'MegaDownloader',
    'DataFetcher',
    'get_data_home',
    'set_data_home',
    'clear_data_home',
]




import numpy as np
import pandas as pd
from pyadlml.constants import ACTIVITY, BOOL, CAT, DEVICE, END_TIME, START_TIME, TIME, VALUE
from pyadlml.dataset.util import get_first_states, get_last_states, infer_dtypes, str_to_timestamp


def remove_days(df_devices, df_activities, days=[], offsets=[], shift='right', retain_corrections=False):
    """ Removes the given days from activities and devices and shifts the succeeding days by that amount
        forward.

    Parameters
    ----------
    df_devices : pd.DataFrame
    df_activities : pd.DataFrame
    days : list
        List of strings
    offsets : list
        Offsets that are added to the corresponding days.
    shift : str, one of ['right', 'left']
        The set of devices/activities that is shifted. Is either the set 'right' to 
        the removed day, meaning days from the future are shifted to the past. Contrary,
        the set 'left' from the day shifts the days from left to right from the past into 
        the future.


    Returns
        df_devices : pd.DataFrame
        df_activities : pd.DataFrame
    """
    df_devs = df_devices.copy()
    df_acts = df_activities.copy()

    assert shift in ['right', 'left']

    # Add offsets to the days specified
    for i in range(len(days)):
        days[i] = str_to_timestamp(days[i])
        if i < len(offsets):
            days[i] = days[i] + pd.Timedelta(offsets[i])

    if shift == 'right':
        # Sort days from last to first
        days = np.array(days)[np.flip(np.argsort(days))]
    else: 
        # Sort days from first to last
        days = np.array(days)[np.argsort(days)]
    dtypes = infer_dtypes(df_devs)

    from pyadlml.plot import plotly_activities_and_devices

    # 1. remove iteratively the latest day and shift the succeeding part accordingly
    for day_lwr_bnd in days:

        # when day is i.e. 2008-03.23 00:00:00 then day after will be 2008-03-24 00:00:00
        # these variables have to be used as only timepoints can be compared as seen below
        day_uppr_bnd = day_lwr_bnd + pd.Timedelta('1D')

        # Remove devices events within the selected day
        mask_inside_day = (day_lwr_bnd < df_devs[TIME]) \
                        & (df_devs[TIME] < day_uppr_bnd)
        removed_devs = df_devs[mask_inside_day].copy()
        df_devs = df_devs[~mask_inside_day]

        # shift the succeeding days timeindex one day in the past
        preceeding_days = (df_devs[TIME] < day_lwr_bnd)
        succeeding_days = (day_uppr_bnd < df_devs[TIME])
        preceeding_devs = df_devs[preceeding_days].copy()
        succeeding_devs = df_devs[succeeding_days].copy()

        if shift == 'right':
            df_devs.loc[succeeding_days, TIME] = df_devs[TIME] - pd.Timedelta('1D')
        else:
            df_devs.loc[preceeding_days, TIME] = df_devs[TIME] + pd.Timedelta('1D')


        # Get first device states from after and last device states from earlier
        dev2last_state = get_last_states(preceeding_devs)
        dev2last_removed_state = get_last_states(removed_devs)
        events_to_add = []
        eps = pd.Timedelta('1ms')
        midnight_corr_time = day_lwr_bnd if shift == 'right' else day_uppr_bnd
        for dev, val in dev2last_state.items():
            def last_states_are_not_equal(dev):
                try:
                    return dev2last_state[dev] != dev2last_removed_state[dev]
                except KeyError:
                    # The case when no events are removed for that device
                    # => no event appeared in the remove d time frame
                    # => the states are equal
                    return False
            if (dev in dtypes[CAT]+dtypes[BOOL]) and last_states_are_not_equal(dev):
                events_to_add.append([midnight_corr_time+eps, dev, dev2last_removed_state[dev]])
                print(f'Warning: Added category {dev2last_removed_state[dev]} device {dev} correction at to midnight: {midnight_corr_time+eps}')
                eps += pd.Timedelta('1ms')
        
        df_devs = pd.concat([df_devs, 
                  pd.DataFrame(columns=[TIME, DEVICE, VALUE], data=events_to_add)])



        # Remove activities in that day that do not extend from the previous day into the selected
        # day or extend from the selected day into the next day
        mask_act_within_day = (day_lwr_bnd < df_acts[START_TIME]) & (df_acts[END_TIME] < day_uppr_bnd)
        df_acts = df_acts[~mask_act_within_day]

        # Special case where one activity starts before the selected day and ends after the selected day
        mask_special = (df_acts[START_TIME] < day_lwr_bnd) & (day_uppr_bnd < df_acts[END_TIME])
        if mask_special.any():
            if shift == 'right':
                df_acts.loc[mask_special, END_TIME] = df_acts[END_TIME] - pd.Timedelta('1D')
            else:
                df_acts.loc[mask_special, START_TIME] = df_acts[START_TIME] + pd.Timedelta('1D')


        df_acts['shifted'] = False
        if shift == 'right':
            # Shift Activities that start in or after the selected day by one day
            succeeding_days = (day_lwr_bnd <= df_acts[START_TIME]) & (day_uppr_bnd < df_acts[END_TIME])
            df_acts.loc[succeeding_days, START_TIME] = df_acts[START_TIME] - pd.Timedelta('1D')
            df_acts.loc[succeeding_days, END_TIME] = df_acts[END_TIME] - pd.Timedelta('1D')
            df_acts.loc[succeeding_days, 'shifted'] = True
        else:
            # Shift Activities that end in or prior the selected day by one day
            preceeding_days = (df_acts[START_TIME] <= day_lwr_bnd) & (df_acts[END_TIME] < day_uppr_bnd)
            df_acts.loc[preceeding_days, START_TIME] = df_acts[START_TIME] + pd.Timedelta('1D')
            df_acts.loc[preceeding_days, END_TIME] = df_acts[END_TIME] + pd.Timedelta('1D')
            df_acts.loc[preceeding_days, 'shifted'] = True

        # Special case where one activity starts before the selected day and ends inside the selected day
        # and there is no activity after the selected day
        #  | day_before | Sel day      | day after
        #     |-------------|
        #    I can't just move the ending one day before as this would reverse START_TIME and END_TIME order
        # -> The last activity ending is clipped to the start of the selected day
        # TODO has to be done before any activity is shifted
        #mask_last_true = pd.Series(np.zeros(len(df_acts), dtype=np.bool_))
        #mask_last_true.iat[-1] = True
        #mask_special = (df_acts[START_TIME] < day) & (df_acts[END_TIME] <= day_after) & mask_last_true
        #if mask_special.any():
        #    assert mask_special.sum() == 1
        #    df_acts.loc[mask_special, END_TIME] = day
        if shift == 'left':
            day_lwr_bnd += pd.Timedelta('1D')
            day_uppr_bnd += pd.Timedelta('1D')

        df_acts = df_acts.sort_values(by=START_TIME).reset_index(drop=True)
        # Merge activities from the day_before that overlap with the shifted activities from day after
        # there are 4 cases where overlaps into the respective days have to be handled
        #                           |   db  |   sd  |   da  |=>|   db  |   sd  |   da  |=>|   db  |   sd  |   da  |
        # 1. db overlaps da            |--------|    |~|          |--------|                 |-----|~|
        #                                                              |~|
        # 2. db intersect into  da      |------|     |~~~~|         |------|                   |----|~~~~|
        #                                                               |~~~~|
        # 3. da overlaps db             |-|   |~~~~~~~|            |-|                       |-|~~~|
        #                                                       |~~~~~~~|
        # 4. da intersect into db       |---|  |~~~~~|          |---|                        |---|~~~|
        #                                                         |~~~~~|
        # 5. da intersect into db         |----||~~~~~|            |-----|                 |---|~~~|
        #                                                      |~~~~~~|
        # case 1:
        # select activities that cross the boundary between days
        mask_db_into_sd = (df_acts[START_TIME] < day_lwr_bnd) \
                        & (df_acts[END_TIME] > day_lwr_bnd) \
                        & (df_acts[END_TIME] < day_uppr_bnd)
        idxs = np.where(mask_db_into_sd)[0]
        assert len(idxs) <= 2
        if len(idxs) == 2:
            if df_acts.iat[idxs[0], 3]:
                # first idxs is shifted activity
                idx_shifted, idx_other = idxs[0], idxs[1]
            else:
                # second idxs is shifted activity
                idx_shifted, idx_other = idxs[1],idxs[0]

            # case 5: case when both activities cross boundaries in each days
            if df_acts.at[idx_shifted, ACTIVITY] == df_acts.at[idx_other, ACTIVITY]:
                # if activities are the same, join them and remove second activity
                # When activity is shifted from past, replace its end_time with the end_time 
                # from the unmoved otherwise the start_time
                replace_time = 1 if shift == 'right' else 0
                df_acts.iat[idx_shifted, replace_time] = df_acts.iat[idx_other, replace_time]
                df_acts = df_acts.drop(idx_other)
            else:
                # Clip  both to midnight [!]
                if shift == 'right':
                    # TODO refactor just swap indices (is it not verbose enough??)
                    df_acts.iat[idx_shifted, 0] = day_lwr_bnd
                    df_acts.iat[idx_other, 1] = day_lwr_bnd + pd.Timedelta('1ms')
                else:
                    df_acts.iat[idx_shifted, 1] = day_lwr_bnd + pd.Timedelta('1ms')
                    df_acts.iat[idx_other, 0] = day_lwr_bnd 


        elif len(idxs) == 1:
            # TODO, do i want this??? or only clip at midnight and not extending into the different days
            idx_overlapping = idxs[0]

            # Check if the overlapping activity is part of the shifted or not
            if df_acts.iat[idx_overlapping, 3]:
                # Case when the shifted activities extend into the day before the selected day
                last_unshifted_act = df_acts.loc[(df_acts[END_TIME] < day_lwr_bnd), :]\
                                .copy().sort_values(by=END_TIME, ascending=True)\
                                .iloc[-1, :]

                # clip extending activities start_time to the end_time of the first shifted activity
                df_acts.iat[idx_overlapping, 0] = last_unshifted_act[END_TIME] + pd.Timedelta('1ms')
            else:
                # Case when the previous activities extends into the selected day
                first_shifted_act = df_acts.loc[(df_acts[START_TIME] > day_lwr_bnd) & (df_acts[END_TIME] < day_uppr_bnd), :]\
                                            .copy().sort_values(by=START_TIME, ascending=True)\
                                            .iloc[0, :]

                # clip extending activities end_time to the start of the first shifted activity
                df_acts.iat[idx_overlapping, 1] = first_shifted_act[START_TIME] - pd.Timedelta('1ms')

        df_acts = df_acts.drop(columns='shifted')

        from pyadlml.dataset._core.devices import correct_devices
        df_devs, corrections_dev = correct_devices(df_devs, retain_corrections)

        #assert not _is_activity_overlapping(df_acts)

    if retain_corrections:
        try:
            corrections_act
        except:
            corrections_act = []
        try:
            corrections_dev
        except:
            corrections_dev = []
        return df_devs, df_acts, corrections_act, corrections_dev
    else:
        return df_devs, df_acts




import sys
import unittest

import pathlib
working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))

from pyadlml.dataset import set_data_home, DEVICE
from pyadlml.preprocessing import StateVectorEncoder

class TestPreprocessingBase(unittest.TestCase):
    __test__ = False


    def __init__(self, *args, **kwargs):
        self.data = None
        self.data_home = None
        self.fetch_method = None

        self.df_activity_attrs = []
        self.lst_activity_attrs = []

        super().__init__(*args, **kwargs)

    def _setUp(self):
        self.data = self.fetch_method(cache=True)

    def setUp(self):
        set_data_home(self.data_home)
        self._setUp()


    def _column_matches_devices(self, df_dev, x):
        l1 = df_dev[DEVICE].unique()
        l2 = list(x.columns[1:])
        return (set(l1) == set(l2)) & (len(l1) == len(l2))


    def test_state_vector_encoder(self):
        from pyadlml.preprocessing import StateVectorEncoder
        df_dev = self.data.df_devices

        # test state vector encoder
        sve = StateVectorEncoder(encode='raw')
        x = sve.fit_transform(df_dev)
        assert len(x) == len(df_dev)
        assert self._column_matches_devices(df_dev, x)
        # TODO test that all devices have same dtype as before the transformation

        sve = StateVectorEncoder(encode='changepoint')
        x = sve.fit_transform(df_dev)
        assert len(x) == len(self.data.df_devices)
        assert self._column_matches_devices(df_dev, x)
        # TODO test for dtypes x_i \in {0,1}

        sve = StateVectorEncoder(encode='last_fired')
        x = sve.fit_transform(df_dev)
        assert len(x) == len(self.data.df_devices)
        assert self._column_matches_devices(df_dev, x)
        # TODO test for dtypes x_i \in {0,1}


    def test_sve_timeslices(self):
        from pyadlml.preprocessing import StateVectorEncoder
        df_dev = self.data.df_devices

        t_res = '1h'
        # test state vector encoder
        sve = StateVectorEncoder(encode='raw', dt=t_res)
        x = sve.fit_transform(df_dev)

        sve = StateVectorEncoder(encode='changepoint', dt=t_res)
        x = sve.fit_transform(df_dev)

        sve = StateVectorEncoder(encode='last_fired', dt=t_res)
        x = sve.fit_transform(df_dev)

        t_res = '30s'
        # test state vector encoder
        sve = StateVectorEncoder(encode='raw', dt=t_res)
        x = sve.fit_transform(df_dev)

        sve = StateVectorEncoder(encode='changepoint', dt=t_res)
        x = sve.fit_transform(df_dev)

        sve = StateVectorEncoder(encode='last_fired', dt=t_res)
        x = sve.fit_transform(df_dev)


import numpy as np



class Explanator(object):
    def __init__(self, mdl, test_x, test_z):
        from hassbrain_algorithm.benchmark.interpretation import ModelWrapper
        wrapped_model = ModelWrapper(mdl)
        class_names = mdl.get_state_lbl_lst()
        feature_names = mdl.get_obs_lbl_lst()

        cat_idxs = [i for i in range(len(feature_names))]
        categorical_names = {}
        for i in cat_idxs:
            categorical_names[i] = {}
            categorical_names[i][0] = "off"
            categorical_names[i][1] = "on"

        from skater.core.local_interpretation.lime.lime_tabular import LimeTabularExplainer
        self._exp = LimeTabularExplainer(
            test_x,
            mode='classification',
            training_labels=test_z,
            feature_names=feature_names,
            categorical_features=cat_idxs,
            categorical_names=categorical_names,
            class_names=class_names)

    def get_explanator(self):
        return self._exp

    def explain(self, x):
        from hassbrain_algorithm.benchmark.interpretation import ModelWrapper

        assert isinstance(x, np.ndarray)
        assert len(x.shape) ==1
        model = ModelWrapper(self)
        lst = self._exp.explain_instance(
            x,
            model.predict_proba).as_list()
        return lst



    def plot_explanation(self, x):
        from hassbrain_algorithm.benchmark.interpretation import ModelWrapper

        assert isinstance(x, np.ndarray)
        model = ModelWrapper(self)
        fig = self._exp.explain_instance(
            x,
            model.predict_proba).as_pyplot_figure()
        fig.show()

    def plot_and_save_explanation(self, x, labels, file_paths):
        from hassbrain_algorithm.benchmark.interpretation import ModelWrapper

        assert isinstance(x, np.ndarray)
        model = ModelWrapper(self)
        enc_labels = self.expl_to_ids(labels)
        exp = self._exp.explain_instance(
            x,
            model.predict_proba,
            labels=enc_labels) #type: Explanation
        for lbl, file_path in zip(enc_labels, file_paths):
            fig = self.as_pyplot_figure(exp, label=lbl)
            import matplotlib.pyplot as plt
            plt.tight_layout()
            fig.savefig(file_path, dpi=fig.dpi)

    def expl_to_ids(self, labels):
        """
        Parameters
        ----------
        labels : list
            list of labels
        Returns
        -------
        list
            encoded labels
        """
        tmp2 = self._exp # type: LimeTabularExplainer
        classnames = tmp2.class_names
        enc_lbl_lst = []
        for lbl in labels:
            assert lbl in classnames
            for i, item in enumerate(classnames):
                if item == lbl:
                    enc_lbl_lst.append(i)
        return enc_lbl_lst


    def as_pyplot_figure(self, exp, label=1, **kwargs):
        """Returns the explanation as a pyplot figure.

        Will throw an error if you don't have matplotlib installed
        Args:
            label: desired label. If you ask for a label for which an
                   explanation wasn't computed, will throw an exception.
                   Will be ignored for regression explanations.
            kwargs: keyword arguments, passed to domain_mapper
        Returns:
            pyplot figure (barchart).
        """
        import matplotlib.pyplot as plt
        explst = exp.as_list(label=label, **kwargs)
        fig = plt.figure()
        vals = [x[1] for x in explst]
        names = [x[0] for x in explst]
        vals.reverse()
        names.reverse()
        colors = ['black' if x > 0 else 'red' for x in vals]
        pos = np.arange(len(explst)) + .5
        plt.barh(pos, vals, align='center', color=colors)
        plt.yticks(pos, names)
        #title = 'Local explanation for class %s' % exp.class_names[label]
        #plt.title(title)
        return fig
